# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""A hierarchical directory of named addresses, expressions of interest and matches.

.
"""
__docformat__ = 'restructuredtext'

import os
import sys
from datetime import datetime
import uuid
import re
from .get_local_ip import get_local_ip

from .general_purpose import *
from .command_line import *
from .ip_networking import *
from .virtual_memory import *
from .convert_memory import *
from .message_memory import *
from .convert_type import *
from .virtual_runtime import *
from .point_runtime import *
from .virtual_point import *
from .point_machine import *
from .bind_type import *
from .listen_connect import *
from .get_response import *

__all__ = [
	'scope_type',
	'DIRECTORY_PORT',
	'DIRECTORY_AT_HOST',
	'DIRECTORY_AT_LAN',
	'directory_at_lan',
	'ConnectTo',
	'AcceptAt',
	'PublishAs',
	'SubscribeTo',
	'Published',
	'Subscribed',
	'ClearPublished',
	'ClearSubscribed',
	'PublishedCleared',
	'SubscribedCleared',
	'NotPublished',
	'NotSubscribed',
	'ObjectDirectory',
	'Available',
	'Delivered',
	'Dropped',
]

#
scope_type = def_type(ScopeOfDirectory)

# Time required for request/response sequence across peer connection.
COMPLETE_A_LOOP = 3.0

# Activate/update the directory-to-directory connections. See
# also ObjectDirectory_READY_Enquiry (support for library processes).
class ConnectTo(object):
	def __init__(self, ipp: HostPort=None):
		self.ipp = ipp or HostPort()

class AcceptAt(object):
	def __init__(self, ipp: HostPort=None):
		self.ipp = ipp or HostPort()

bind(ConnectTo)
bind(AcceptAt)

# Declare publish/subscribe within the host process.
# Sent by the application objects.
class PublishAs(object):
	def __init__(self, name: str=None, scope: ScopeOfDirectory=None, publisher_address: Address=None, encrypted: bool=None):
		self.name = name
		self.scope = scope
		self.publisher_address = publisher_address
		self.encrypted = encrypted

class SubscribeTo(object):
	def __init__(self, search: str=None, scope: ScopeOfDirectory=None, subscriber_address: Address=None):
		self.search = search
		self.scope = scope
		self.subscriber_address = subscriber_address

bind(PublishAs)
bind(SubscribeTo)

# Internal record of pub/subs held within each directory in the tree.
# Also sent to subscribers/publishers as confirmation of sub/pub and
# expected as the arg to clear_published, et al.
# Propagated up the hierarchy.
class Published(object):
	"""
	Successful completion of :func:`~.publish`.

	:param name: name the service is known by
	:param scope: scope in which to search
	:param published_id: unique identity assigned to registration
	:param listening_ipp: IP and port assigned to this service
	:param home_address: address of the publishing process
	"""
	def __init__(self, name: str=None, scope: ScopeOfDirectory=None, encrypted: bool=False,
			published_id: UUID=None, listening_ipp: HostPort=None,
			home_address: Address=None):
		self.name = name
		self.scope = scope
		self.encrypted = encrypted
		self.published_id = published_id
		self.listening_ipp = listening_ipp or HostPort()
		self.home_address = home_address

class Subscribed(object):
	"""
	Successful completion of :func:`~.subscribe`.

	:param search: pattern to match with services
	:param scope: scope in which to search
	:param subscribed_id: unique identity assigned to registration
	:param home_address: address of the subscribing process
	"""
	def __init__(self, search: str=None, scope: ScopeOfDirectory=None,
			subscribed_id: UUID=None, home_address: Address=None):
		self.search = search
		self.scope = scope
		self.subscribed_id = subscribed_id
		self.home_address = home_address

bind(Published)
bind(Subscribed)

class Advisory(object):
	"""A warning to the receiving directory, that there is a collision at a higher level."""
	def __init__(self, name: str=None, scope: ScopeOfDirectory=None, published_id: UUID=None):
		self.name = name
		self.scope = scope
		self.published_id = published_id

bind(Advisory)

# Instructions and confirmations from the application to retract
# the specified pub/sub.
class ClearPublished(object):
	def __init__(self, name: str=None, scope: ScopeOfDirectory=None, published_id: UUID=None, note: str=None):
		self.name = name
		self.scope = scope
		self.published_id = published_id
		self.note = note

class ClearSubscribed(object):
	def __init__(self, search: str=None, scope: ScopeOfDirectory=None, subscribed_id: UUID=None, note: str=None):
		self.search = search
		self.scope = scope
		self.subscribed_id = subscribed_id
		self.note = note

class PublishedCleared(object):
	"""
	Successful completion of :func:`~.clear_published`.

	:param name: name the service is known by
	:param scope: scope in which to search
	:param published_id: unique identity assigned to registration
	:param note: short description appearing in logs
	"""
	def __init__(self, name: str=None, scope: ScopeOfDirectory=None, published_id: UUID=None, note: str=None):
		self.name = name
		self.scope = scope
		self.published_id = published_id
		self.note = note

class SubscribedCleared(object):
	"""
	Successful completion of :func:`~.clear_subscribed`.

	:param name: name the service is known by
	:param scope: scope in which to search
	:param subscribed_id: unique identity assigned to registration
	:param note: short description appearing in logs
	"""
	def __init__(self, search: str=None, scope: ScopeOfDirectory=None, subscribed_id: UUID=None, note: str=None):
		self.search = search
		self.scope = scope
		self.subscribed_id = subscribed_id
		self.note = note

bind(ClearPublished)
bind(ClearSubscribed)
bind(PublishedCleared)
bind(SubscribedCleared)

# When pub/sub fails.
class NotPublished(Faulted):
	"""
	Unsuccessful completion of :func:`~.publish`.

	Derived from :class:`~.Faulted`

	:param name: name the service would be known by
	:param scope: scope in which service would be available
	:param note: short description added to logs
	"""
	def __init__(self, name: str=None, scope: ScopeOfDirectory=None, note: str=None):
		self.name = name
		self.scope = scope
		self.note = note
		Faulted.__init__(self,f'cannot publish as "{name}"', note)

class NotSubscribed(Faulted):
	"""
	Unsuccessful completion of :func:`~.subscribe`.

	Derived from :class:`~.Faulted`

	:param name: pattern for matching services
	:param scope: scope in which to search
	:param note: short description added to logs
	"""
	def __init__(self, search: str=None, scope: ScopeOfDirectory=None, note: str=None):
		self.search = search
		self.scope = scope
		self.note = note
		Faulted.__init__(self,f'cannot subscribe to "{search}"', note)

bind(NotPublished, explanation=str, error_code=int, exit_status=int)
bind(NotSubscribed, explanation=str, error_code=int, exit_status=int)

# Bulk transfer of listings between directories, e.g. on
# reconnect to parent.
class PublishedDirectory(object):
	def __init__(self, published: list[Published]=None, subscribed: list[Subscribed]=None):
		self.published = published or []
		self.subscribed = subscribed or []

bind(PublishedDirectory)

# When an accepted directory is lost, all the pub/sub listings
# originating from that connection are cleared out of the hierarchy.
class ClearListings(object):
	def __init__(self, subscribers: set[UUID]=None, publishers: set[UUID]=None):
		self.subscribers = subscribers or set()
		self.publishers = publishers or set()

bind(ClearListings)

# Custom message from route to loadable library process.
class OpenLibrary(object):
	def __init__(self, published_id: UUID=None, subscribed_id: UUID=None):
		self.published_id = published_id
		self.subscribed_id = subscribed_id

bind(OpenLibrary)

# Messages from route to pub/sub home processes to inform the
# receivers of another routing option.
# Base class for all.
class SubscriberRoute(object):
	def __init__(self, route_id: UUID=None, scope: ScopeOfDirectory=None,
			subscribed_id: UUID=None, published_id: UUID=None,
			name: str=None):
		self.route_id = route_id
		self.scope = scope
		self.subscribed_id = subscribed_id
		self.published_id = published_id
		self.name = name

# Derived class for RouteOverConnect.
class RouteOverLoop(SubscriberRoute):
	def __init__(self, route_id: UUID=None, scope: ScopeOfDirectory=None, encrypted: bool=False,
			subscribed_id: UUID=None, published_id: UUID=None,
			ipp: HostPort=None, name: str=None):
		# Base members.
		self.route_id = route_id
		self.scope = scope
		self.encrypted = encrypted
		self.subscribed_id = subscribed_id
		self.published_id = published_id
		self.name = name
		# Specialized members.
		self.ipp = ipp

# Derived class for RouteInProcess.
class RouteToAddress(SubscriberRoute):
	def __init__(self, route_id: UUID=None, scope: ScopeOfDirectory=None,
			subscribed_id: UUID=None, published_id: UUID=None,
			subscriber_address: Address=None, publisher_address: Address=None,
			name: str=None, opened_at: datetime=None):
		# Base members.
		self.route_id = route_id
		self.scope = scope
		self.subscribed_id = subscribed_id
		self.published_id = published_id
		self.subscriber_address = subscriber_address
		self.publisher_address = publisher_address
		self.name = name
		self.opened_at = opened_at

# Messages from route to pub/sub home processes, to delete the given route.
class ClearSubscriberRoute(object):
	def __init__(self, subscribed_id: UUID=None, name: str=None, route_id: UUID=None):
		self.subscribed_id = subscribed_id
		self.name = name
		self.route_id = route_id

class ClearPublisherRoute(object):
	def __init__(self, published_id: UUID=None, name: str=None, route_id: UUID=None):
		self.published_id = published_id
		self.name = name
		self.route_id = route_id

bind(SubscriberRoute)
bind(RouteOverLoop)
bind(RouteToAddress)
bind(ClearSubscriberRoute)
bind(ClearPublisherRoute)

# From directory to connector. Pass the baton for establishing
# communications.
class RequestLoop(object):
	def __init__(self, name: str=None, scope: ScopeOfDirectory=None, route_id: UUID=None,
			subscribed_id: UUID=None, published_id: UUID=None,
			subscriber_address: Address=None):
		self.name = name
		self.scope = scope
		self.route_id = route_id
		self.subscribed_id = subscribed_id
		self.published_id = published_id
		self.subscriber_address = subscriber_address

# From ConnectToPeer to ListenForPeer. Establish
# a virtual circuit.
class OpenLoop(object):
	def __init__(self, name: str=None, scope: ScopeOfDirectory=None, route_id: UUID=None, subscribed_id: UUID=None, published_id: UUID=None, subscriber_address: Address=None):
		self.name = name
		self.scope = scope
		self.route_id = route_id
		self.subscribed_id = subscribed_id
		self.published_id = published_id
		self.subscriber_address = subscriber_address

# Response to OpenLoop.
class LoopOpened(object):
	def __init__(self, publisher_address: Address=None):
		self.publisher_address = publisher_address

# From ConnectToPeer to ListenForPeer. Clear the
# virtual cicuit.
class CloseLoop(object):
	def __init__(self, name: str=None, scope: ScopeOfDirectory=None, route_id: UUID=None, subscribed_id: UUID=None, published_id: UUID=None, subscriber_address: Address=None):
		self.name = name
		self.scope = scope
		self.route_id = route_id
		self.subscribed_id = subscribed_id
		self.published_id = published_id
		self.subscriber_address = subscriber_address

# Response to CloseLoop.
class LoopClosed(object):
	def __init__(self, publisher_address: Address=None):
		self.publisher_address = publisher_address

# From directory to connector. Subscriber is rerouting,
# e.g. upgrading.
class DropLoop(object):
	def __init__(self, name: str=None, scope: ScopeOfDirectory=None, route_id: UUID=None, subscribed_id: UUID=None, published_id: UUID=None, subscriber_address: Address=None):
		self.name = name
		self.scope = scope
		self.route_id = route_id
		self.subscribed_id = subscribed_id
		self.published_id = published_id
		self.subscriber_address = subscriber_address

# Response to DropLoop and also when remote object
# clears a pub/sub.
class LoopDropped(object):
	def __init__(self, subscribed_id: UUID=None, name: str=None, route_id: UUID=None):
		self.subscribed_id = subscribed_id
		self.name = name
		self.route_id = route_id

bind(RequestLoop)
bind(OpenLoop)
bind(LoopOpened)
bind(CloseLoop)
bind(LoopClosed)
bind(DropLoop)
bind(LoopDropped)

# Notifications from directory to pub/sub regarding presence of virtual circuit.
# A publisher is available at self.return_address.
class Available(object):
	"""Session notification, peer transport to service established.

	:param name: name of the matched service
	:param scope: scope at which name was matched
	:param route_id: unique identity assigned to match
	:param subscribed_id: unique identity assigned to subscription
	:param published_id: unique identity assigned to publication
	:param publisher_address: address of the publishing process
	:param opened_at: moment the match occurred
	"""
	def __init__(self, name: str=None, scope: ScopeOfDirectory=None, route_id: UUID=None,
			subscribed_id: UUID=None, published_id: UUID=None,
			publisher_address: Address=None, opened_at: datetime=None):
		self.name = name
		self.scope = scope
		self.route_id = route_id
		self.subscribed_id = subscribed_id
		self.published_id = published_id
		self.publisher_address = publisher_address
		self.opened_at = opened_at

# A subscriber is at self.return_address.
class Delivered(object):
	"""Session notification, peer transport to subscriber established.

	:param name: name of the matched service
	:param scope: scope at which name was matched
	:param route_id: unique identity assigned to match
	:param subscribed_id: unique identity assigned to subscription
	:param published_id: unique identity assigned to publication
	:param subscriber_address: address of the subscribing process
	:param opened_at: moment the match occurred
	"""
	def __init__(self, name: str=None, scope: ScopeOfDirectory=None, route_id: UUID=None,
			subscribed_id: UUID=None, published_id: UUID=None,
			subscriber_address: Address=None, opened_at: datetime=None):
		self.name = name
		self.scope = scope
		self.route_id = route_id
		self.subscribed_id = subscribed_id
		self.published_id = published_id
		self.subscriber_address = subscriber_address
		self.opened_at = opened_at

# An existing circuit has been cleared.
class Dropped(object):
	"""Session notification, peer transport to published/subscriber lost.

	:param name: name of the matched service
	:param scope: scope at which name was matched
	:param route_id: unique identity assigned to match
	:param subscribed_id: unique identity assigned to subscription
	:param published_id: unique identity assigned to publication
	:param remote_address: address of the subscriber/publisher
	:param opened_at: moment the match occurred
	"""
	def __init__(self, name: str=None, scope: ScopeOfDirectory=None, route_id: UUID=None,
			subscribed_id: UUID=None, published_id: UUID=None,
			remote_address: Address=None, opened_at: datetime=None):
		self.name = name
		self.scope = scope
		self.route_id = route_id
		self.subscribed_id = subscribed_id
		self.published_id = published_id
		self.remote_address = remote_address
		self.opened_at = opened_at

bind(Available)
bind(Delivered)
bind(Dropped)

#
#
class OpenDirectory(object):
	def __init__(self, scope: ScopeOfDirectory=None):
		self.scope = scope or ScopeOfDirectory.HOST

class ListDirectory(object): pass
class GetDirectory(object): pass
class DirectoryOpened(object): pass

class DirectoryRoute(object):
	"""A [scope][name] route from subscribed id to published id.

	.
	"""
	def __init__(self, name: str=None,
			subscribed_id: UUID=None, published_id: UUID=None):
		self.name = name
		self.subscribed_id = subscribed_id
		self.published_id = published_id

class PeerSession(object):
	def __init__(self, name: str=None, route: Any=None, status: Any=None):
		self.name = name
		self.route = route
		self.status = status

class DirectoryPeer(object):
	def __init__(self, subscribed_id: UUID=None, search: str=None, session: list[PeerSession]=None):
		self.subscribed_id = subscribed_id
		self.search = search
		self.session = session or []

class DirectoryListing(object):
	"""Network administration, contents.

	.
	"""
	def __init__(self, unique_id: UUID=None, executable: str=None,
			directory_address: Address=None, scope: ScopeOfDirectory=None,
			primary_ip: str=None,
			connect_to_directory: HostPort=None, accept_directories_at: HostPort=None,
			published: list[Published]=None, subscribed: list[Subscribed]=None,
			routed: list[DirectoryRoute]=None,
			peer: dict[UUID, DirectoryPeer]=None,
			sub_directory: dict[HostPort, Address]=None):
		self.unique_id = unique_id
		self.executable = executable
		self.directory_address = directory_address
		self.scope = scope
		self.primary_ip = primary_ip or '(not detected)'
		self.connect_to_directory = connect_to_directory
		self.accept_directories_at = accept_directories_at
		self.published = published or []
		self.subscribed = subscribed or []
		self.routed = routed or []
		self.peer = peer or {}
		self.sub_directory = sub_directory or {}

bind(OpenDirectory)
bind(DirectoryOpened)
bind(ListDirectory)
bind(GetDirectory)
bind(DirectoryRoute)
bind(PeerSession)
bind(DirectoryPeer)
bind(DirectoryListing)

#
class INITIAL: pass
class PENDING: pass
class READY: pass
class OPENING: pass

DIRECTORY_PORT			= 54195
DIRECTORY_AT_EPHEMERAL	= HostPort('127.0.0.1', 0)
DIRECTORY_AT_HOST		= HostPort('127.0.0.1', DIRECTORY_PORT)
DIRECTORY_AT_LAN		= HostPort('192.168.0.195', DIRECTORY_PORT)

# A managed listen. One required for every publish beyond
# the process scope.
class ListeningForPeer(Point, StateMachine):
	def __init__(self, name: str=None, scope: ScopeOfDirectory=None, address: Address=None, encrypted: bool=False):
		Point.__init__(self)
		StateMachine.__init__(self, INITIAL)
		self.name = name
		self.scope = scope
		self.address = address
		self.encrypted = encrypted
		self.listening = None
		self.accepted = {}

def ListeningForPeer_INITIAL_Start(self, message):
	# Tune the listen address according to the scope of the publish. Host
	# portion is overruled where the publish is listed in higher scopes,
	# e.g. LAN. All use ephemeral ports.
	if self.scope.value < ScopeOfDirectory.HOST.value:
		ipp = HostPort('0.0.0.0', 0)

	elif self.scope.value < ScopeOfDirectory.PROCESS.value:
		ipp = HostPort('127.0.0.1', 0)

	else:
		self.complete(Faulted(f'Cannot peer for scope [{self.scope}]'))

	listen(self, ipp, encrypted=self.encrypted)
	return PENDING

def ListeningForPeer_PENDING_Listening(self, message):
	self.listening = message
	self.send(message.listening_ipp, self.parent_address)
	return READY

def ListeningForPeer_PENDING_NotListening(self, message):
	self.complete(message)

def ListeningForPeer_READY_Accepted(self, message):
	# Tracking of open loops.
	self.accepted[self.return_address[-1]] = {}
	return READY

def ListeningForPeer_READY_Closed(self, message):
	p = self.accepted.pop(self.return_address[-1], None)
	# Going down. Send out the end-of-session notifications
	# to those that received the start-of-session.
	if p is not None:
		for k, v in p.items():
			a, opened_at = v
			d = Dropped(name=a.name, scope=a.scope, route_id=a.route_id,
				subscribed_id=a.subscribed_id, published_id=a.published_id,
				remote_address=k, opened_at=opened_at)
			self.forward(d, self.address, k)
	return READY

def ListeningForPeer_READY_OpenLoop(self, message):
	opened_at = world_now()
	self.accepted[self.return_address[-1]][message.subscriber_address] = [message, opened_at]

	self.reply(LoopOpened(publisher_address=self.address))

	d = Delivered(name=message.name, scope=message.scope, route_id=message.route_id,
		subscribed_id=message.subscribed_id, published_id=message.published_id,
		subscriber_address=message.subscriber_address, opened_at=opened_at)

	self.forward(d, self.address, message.subscriber_address)
	return READY

def ListeningForPeer_READY_CloseLoop(self, message):
	p, opened_at = self.accepted[self.return_address[-1]].pop(message.subscriber_address, None)

	self.reply(LoopClosed(publisher_address=self.address))

	d = Dropped(name=message.name, scope=message.scope, route_id=message.route_id,
		subscribed_id=message.subscribed_id, published_id=message.published_id,
		remote_address=message.subscriber_address, opened_at=opened_at)

	self.forward(d, self.address, message.subscriber_address)
	return READY

def ListeningForPeer_READY_Stop(self, message):
	self.complete(Aborted())

def ListeningForPeer_READY_NotListening(self, message):
	self.complete(Aborted())

LISTENING_FOR_PEER_DISPATCH = {
	INITIAL: (
		(Start,),
		()
	),
	PENDING: (
		(Listening, NotListening),
		()
	),
	READY: (
		(Accepted, Closed,
		OpenLoop, CloseLoop,
		Stop,
		NotListening),
		()
	),
}

bind(ListeningForPeer, LISTENING_FOR_PEER_DISPATCH)

# A managed connect. One required for every outbound route
# on a unique ip+port.
class ConnectToPeer(Point, StateMachine):
	def __init__(self, ipp: HostPort=None, encrypted: bool=False):
		Point.__init__(self)
		StateMachine.__init__(self, INITIAL)
		self.ipp = ipp
		self.encrypted = encrypted
		self.request = []		# All the RequestLoops.
		self.available = []		# Requests answered by LoopOpened.

	def delete_request(self, route_id):
		d = None
		for i, ra in enumerate(self.request):
			r, a = ra
			if r.route_id == route_id:
				d = i
				break
		if d is not None:
			r = self.request.pop(d)
			return r
		return None

	def delete_available(self, route_id):
		d = None
		for i, a in enumerate(self.available):
			if a[0].route_id == route_id:
				d = i
				break
		if d is not None:
			a = self.available.pop(d)
			return a
		return None

	def not_available(self):
		for a, s, p in self.available:
			d = LoopDropped(subscribed_id=a.subscribed_id, name=a.name, route_id=a.route_id)
			self.send(d, self.parent_address)

			d = Dropped(name=a.name, scope=a.scope, route_id=a.route_id,
				subscribed_id=a.subscribed_id, published_id=a.published_id,
				remote_address=a.publisher_address, opened_at=a.opened_at)

			self.forward(d, s, p)

def ConnectToPeer_INITIAL_Start(self, message):
	localhost = self.ipp.host.startswith('127.')
	keep_alive = not localhost
	connect(self, self.ipp, keep_alive=keep_alive, encrypted=self.encrypted)
	return PENDING

def ConnectToPeer_PENDING_Connected(self, message):
	self.connected = message

	def opened(self, loop, args):
		request = args.request
		client_address = args.client_address
		if not isinstance(loop, LoopOpened):
			self.warning(f'Closing peer connection (unexpected looping response {loop})')
			self.send(Close(), self.connected.proxy_address)
			self.send(loop, client_address)
			return

		a = Available(name=request.name, scope=request.scope, route_id=request.route_id,
			published_id=request.published_id, subscribed_id=request.subscribed_id,
			publisher_address=loop.publisher_address, opened_at=world_now())

		self.forward(a, request.subscriber_address, loop.publisher_address)
		self.available.append((a, request.subscriber_address, loop.publisher_address))
		self.send(self.connected, client_address)

	# Send OpenLoop on behalf of each client.
	for ra in self.request:
		r, a = ra
		address = r.subscriber_address
		open = OpenLoop(name=r.name, scope=r.scope, route_id=r.route_id,
			subscribed_id=r.subscribed_id, published_id=r.published_id,
			subscriber_address=address)
		g = self.create(GetResponse, open, self.connected.proxy_address, seconds=COMPLETE_A_LOOP)
		self.on_return(g, opened, request=r, client_address=a)
	return READY

def ConnectToPeer_PENDING_NotConnected(self, message):
	self.complete()

def ConnectToPeer_PENDING_RequestLoop(self, message):
	request_and_address = (message, self.return_address)
	self.request.append(request_and_address)
	return PENDING

def ConnectToPeer_PENDING_DropLoop(self, message):
	if self.delete_request(message.route_id):
		d = LoopDropped(subscribed_id=message.subscribed_id, name=message.name, route_id=message.route_id)
		self.reply(d)
	return PENDING

def ConnectToPeer_READY_RequestLoop(self, message):
	request_and_address = (message, self.return_address)
	self.request.append(request_and_address)

	def opened(self, loop, args):
		request = args.request
		client_address = args.client_address
		if not isinstance(loop, LoopOpened):
			self.warning(f'Closing peer connection (unexpected looping response {loop})')
			self.send(Close(), self.connected.proxy_address)
			self.send(loop, self.client_address)
			return

		a = Available(name=request.name, scope=request.scope, route_id=request.route_id,
			published_id=request.published_id, subscribed_id=request.subscribed_id,
			publisher_address=loop.publisher_address, opened_at=world_now())

		self.forward(a, request.subscriber_address, loop.publisher_address)
		self.available.append((a, request.subscriber_address, loop.publisher_address))
		self.send(self.connected, client_address)

	open = OpenLoop(name=message.name, scope=message.scope, route_id=message.route_id,
		subscribed_id=message.subscribed_id, published_id=message.published_id,
		subscriber_address=message.subscriber_address)

	a = self.create(GetResponse, open, self.connected.proxy_address, seconds=COMPLETE_A_LOOP)
	self.on_return(a, opened, request=message, client_address=self.return_address)
	return READY

def ConnectToPeer_READY_DropLoop(self, message):
	dr = self.delete_request(message.route_id)
	da = self.delete_available(message.route_id)
	if dr is None or da is None:
		self.warning(f'Request to drop unknown/incomplete loop')
		d = LoopDropped(subscribed_id=message.subscribed_id, name=message.name, route_id=message.route_id)
		self.reply(d)
		return READY

	def closed(self, loop, args):
		request, available, return_address = args.request, args.available, args.return_address
		client_address = args.client_address

		if isinstance(loop, LoopClosed):
			d = Dropped(name=request.name, scope=request.scope, route_id=request.route_id,
				subscribed_id=request.subscribed_id, published_id=request.published_id,
				remote_address=available[2], opened_at=available[0].opened_at)
			self.forward(d, available[1], available[2])

			d = LoopDropped(subscribed_id=request.subscribed_id, name=request.name, route_id=request.route_id)
			self.send(d, return_address)

			if len(self.request) == 0:
				self.start(T1, GRACE_BEFORE_CLEARANCE)
			return

		if isinstance(self.connected, Connected):
			self.send(Close(), self.connected.proxy_address)
		self.complete()

	address = message.subscriber_address
	close = CloseLoop(name=message.name, scope=message.scope, route_id=message.route_id,
		subscribed_id=message.subscribed_id, published_id=message.published_id,
		subscriber_address=address)

	a = self.create(GetResponse, close, self.connected.proxy_address, seconds=COMPLETE_A_LOOP)
	self.on_return(a, closed, request=dr[0], client_address=dr[1], available=da, return_address=self.return_address)
	return READY

def ConnectToPeer_READY_T1(self, message):
	if len(self.request) == 0:
		self.send(Close(), self.connected.proxy_address)
	return READY

def ConnectToPeer_READY_Returned(self, message):
	d = self.debrief()
	if isinstance(d, OnReturned):
		d(self, message)
	return READY

def ConnectToPeer_READY_Closed(self, message):
	self.not_available()
	self.complete()

def ConnectToPeer_READY_Stop(self, message):
	self.complete()

CONNECT_TO_PEER_DISPATCH = {
	INITIAL: (
		(Start,),
		()
	),
	PENDING: (
		(Connected, NotConnected,
		RequestLoop, DropLoop),
		()
	),
	READY: (
		(RequestLoop, DropLoop,
		T1,
		Returned,
		Closed,
		Stop),
		()
	),
}

bind(ConnectToPeer, CONNECT_TO_PEER_DISPATCH)

# A route is available that would require a peer connection
# from subscriber to publisher.
class RouteOverConnect(Point, StateMachine):
	def __init__(self, route_id: UUID=None, scope: ScopeOfDirectory=None, encrypted: bool=False, subscriber: Subscribed=None, publisher: Published=None):
		Point.__init__(self)
		StateMachine.__init__(self, INITIAL)
		self.route_id = route_id
		self.scope = scope
		self.encrypted = encrypted
		self.subscriber = subscriber
		self.publisher = publisher

def RouteOverConnect_INITIAL_Start(self, message):
	r = RouteOverLoop(route_id=self.route_id, scope=self.scope, encrypted=self.encrypted,
		subscribed_id=self.subscriber.subscribed_id, published_id=self.publisher.published_id,
		name=self.publisher.name, ipp=self.publisher.listening_ipp)

	self.send(r, self.subscriber.home_address)
	return READY

def RouteOverConnect_READY_Stop(self, message):
	s = self.subscriber
	p = self.publisher
	self.send(ClearSubscriberRoute(subscribed_id=s.subscribed_id, name=p.name, route_id=self.route_id), s.home_address)
	self.send(ClearPublisherRoute(published_id=p.published_id, name=p.name, route_id=self.route_id), p.home_address)
	self.complete(Aborted())

ROUTE_OVER_PEER_DISPATCH = {
	INITIAL: (
		(Start,),
		()
	),
	READY: (
		(Stop,),
		()
	),
}

bind(RouteOverConnect, ROUTE_OVER_PEER_DISPATCH)

# A match has been found between two objects within a process.
# This also covers the process-to-library scenario.
class RouteInProcess(Point, StateMachine):
	def __init__(self, route_id: UUID=None, subscriber: Subscribed=None, publisher: Published=None, subscribe_to: SubscribeTo=None, publish_as: PublishAs=None):
		Point.__init__(self)
		StateMachine.__init__(self, INITIAL)
		self.route_id = route_id
		self.subscriber = subscriber
		self.publisher = publisher
		self.subscribe_to = subscribe_to
		self.publish_as = publish_as

	def send_route(self):
		opened_at = world_now()
		r = RouteToAddress(route_id=self.route_id, scope=ScopeOfDirectory.PROCESS,
			subscribed_id=self.subscriber.subscribed_id, published_id=self.publisher.published_id,
			subscriber_address=self.subscribe_to.subscriber_address, publisher_address=self.publish_as.publisher_address,
			name=self.publisher.name, opened_at=opened_at)

		self.send(r, self.subscriber.home_address)

def RouteInProcess_INITIAL_Start(self, message):
	pa = self.publish_as
	st = self.subscribe_to
	if pa is None and st is None:
		self.complete(Faulted('inter-library routing not supported'))

	if pa is None:
		self.send(OpenLibrary(published_id=self.publisher.published_id), self.publisher.home_address)
		return OPENING

	elif st is None:
		self.send(OpenLibrary(subscribed_id=self.subscriber.subscribed_id), self.subscriber.home_address)
		return OPENING

	self.send_route()
	return READY

def RouteInProcess_OPENING_SubscribeTo(self, message):
	self.subscribe_to = message
	self.send_route()
	return READY

def RouteInProcess_OPENING_PublishAs(self, message):
	self.publish_as = message
	self.send_route()
	return READY

def RouteInProcess_OPENING_Stop(self, message):
	self.complete(Aborted())

def RouteInProcess_READY_Stop(self, message):
	#if self.publish_as is None:
	#	self.forward(OpenLibrary(self.publisher.name), self.publisher.home_address, self.subscriber_address)
	#	return READY

	s = self.subscriber
	p = self.publisher
	self.send(ClearSubscriberRoute(subscribed_id=s.subscribed_id, name=p.name, route_id=self.route_id), s.home_address)
	self.send(ClearPublisherRoute(published_id=p.published_id, name=p.name, route_id=self.route_id), p.home_address)
	self.complete(Aborted())

ROUTE_IN_PROCESS_DISPATCH = {
	INITIAL: (
		(Start,),
		()
	),
	OPENING: (
		(SubscribeTo, PublishAs, Stop),
		()
	),
	READY: (
		(Stop,),
		()
	),
}

bind(RouteInProcess, ROUTE_IN_PROCESS_DISPATCH)

#
def find_route(route, routing):
	for r in routing:
		if r.scope == route.scope or r.route_id == route.route_id:
			return True
	return False

def scope_route(route, routing):
	for i, r in enumerate(routing):
		if r.scope == route.scope:
			return i
	return None

def shortest_route(routing, excluding=None):
	best = None
	for r in routing:
		if r.route_id == excluding:
			continue
		if best is None or r.scope.value > best.scope.value:
			best = r
	return best

def add_route(route, routing):
	routing.append(route)
	shortest = shortest_route(routing)
	return shortest

def delete_route(route_id, routing):
	d = None
	for i, r in enumerate(routing):
		if r.route_id == route_id:
			d = i
	if d is not None:
		r = routing.pop(d)
		return r
	return None

four_octets = re.compile(r'([0-9]+)\.([0-9]+)\.([0-9]+)\.([0-9]+)')

def directory_at_lan():
	d = get_local_ip()
	if not d:
		return None

	m = four_octets.fullmatch(d)
	if not m:
		return None

	if d.startswith('192.168.'):
		leading_octets = '192.168.0'
	elif d.startswith('172.'):
		octet_2 = int(m[2])
		if octet_2 < 16 or octet_2 > 31:
			return None
		leading_octets = '172.16.0'
	elif d.startswith('10.'):
		leading_octets = '10.0.0'
	else:
		return None

	a = f'{leading_octets}.195'
	return HostPort(a, DIRECTORY_PORT)

RECONNECT_DELAY = [1.0, 8.0, 32.0, 120.0]	# 1-based indexing from enum, i.e. [0] is not used.
GRACE_BEFORE_CLEARANCE = 8.0

class ObjectDirectory(Threaded, StateMachine):
	def __init__(self, directory_scope: ScopeOfDirectory=None,
			connect_to_directory: HostPort=None, accept_directories_at: HostPort=None,
			encrypted: bool=None):
		Threaded.__init__(self)
		StateMachine.__init__(self, INITIAL)
		self.directory_scope = directory_scope or ScopeOfDirectory.PROCESS
		self.connect_to_directory = connect_to_directory or HostPort()
		self.accept_directories_at = accept_directories_at or HostPort()
		self.encrypted = encrypted

		self.unique_id = uuid.uuid4()
		self.reconnect_delay = None

		# Links to the upper and lower parts of the hierarchy.
		self.connected = None
		self.listening = None
		self.accepted = {}				# Remember who connects from below and what they provide.
		self.pending_enquiry = set()

		# Keep the db clean.
		self.unique_publish = {}
		self.unique_subscribe = {}

		# For quick lookup and distribution.
		self.listed_publisher = {}
		self.listed_subscriber = {}

		# For quick matching.
		self.published_name = {}
		self.subscribed_search = {}

		# Routed listings.
		self.routed_publish = {}
		self.routed_subscribe = {}

		# Routing tables for active loops.
		self.subscriber_routing = {}

		# Connections across group, host and lan domains.
		self.peer_connect = {}

		self.directory_opened = None

	def calculate_reconnect(self, host):
		if host is None:
			return
		s = local_private_other(host)
		seconds = RECONNECT_DELAY[s.value]
		self.reconnect_delay = spread_out(seconds, 10)
		self.trace(f'Update parameter', reconnect_delay=self.reconnect_delay)

	def keep_alive(self):
		host = self.connect_to_directory.host
		localhost = host.startswith('127.')
		return not localhost

	def encrypted_directory(self):
		if self.encrypted is None:
			return CL.encrypted_process
		return self.encrypted

	def auto_connect(self, message):
		# If certain conditions are met, auto-assign a parent directory address
		# and start connecting.
		connecting_to_host = self.connect_to_directory.host is not None
		to_be_pushed = message.scope.value < self.directory_scope.value
		if connecting_to_host or not to_be_pushed:
			return

		if self.directory_scope in (ScopeOfDirectory.PROCESS, ScopeOfDirectory.GROUP):
			self.connect_to_directory = DIRECTORY_AT_HOST
		elif self.directory_scope  == ScopeOfDirectory.HOST:
			a = directory_at_lan()
			if a is None:
				return
			self.connect_to_directory = a
		else:
			return

		connect(self, self.connect_to_directory,
			keep_alive=self.keep_alive(),
			encrypted=self.encrypted_directory())
		self.calculate_reconnect(self.connect_to_directory.host)

	def add_publisher(self, listing: Published, origin: Address, publish: PublishAs):
		name = listing.name
		scope = listing.scope
		encrypted = listing.encrypted

		# Existence - by id.
		lp = self.listed_publisher.get(listing.published_id, None)
		if lp is not None:
			if lp[0].home_address == listing.home_address:
				self.trace(f'Publish ignored "{name}" (refresh)')
			else:
				self.warning(f'Publish ignored "{name}" (same id, multiple addresses)')
			return False

		# And search name.
		lp = self.published_name.get(name, None)
		if lp is not None:
			self.warning(f'Cannot publish "{name}" (multiple ids, same name)')
			return False

		if publish:
			# Publishing object is in this process. Is an
			# access point required.
			if scope.value < ScopeOfDirectory.PROCESS.value:
				a = self.create(ListeningForPeer,
					name=name, scope=scope,
					address=publish.publisher_address, encrypted=encrypted)
				self.assign(a, (listing, publish))
				return False

		elif self.directory_scope == ScopeOfDirectory.LAN:
			# This directory overlooks a LAN. Overwrite the given
			# listening ip (0.0.0.0) with the ip provided by sockets.
			a, sub, pub = self.accepted.get(self.return_address[-1])
			listing.listening_ipp = HostPort(a.opened_ipp.host, listing.listening_ipp.port)

		elif self.directory_scope in (ScopeOfDirectory.HOST, ScopeOfDirectory.GROUP):
			# This directory overlooks a HOST or GROUP. Overwrite whatever
			# was provided with the loopback address.
			listing.listening_ipp = HostPort('127.0.0.1', listing.listening_ipp.port)
		elif self.directory_scope in (ScopeOfDirectory.PROCESS, ScopeOfDirectory.LIBRARY):
			pass
		else:
			self.warning(f'Scope [{self.directory_scope}] not implemented')
			return False

		self.console(f'Published "{name}"[{self.directory_scope}] ({listing.listening_ipp})')

		lp = (listing, publish)
		self.published_name[name] = lp
		self.listed_publisher[listing.published_id] = lp

		if origin is not None:
			origin.add(listing.published_id)
		return True

	def add_subscriber(self, listing: Subscribed, origin: Address, subscribe: SubscribeTo):
		search = listing.search
		scope = listing.scope
		subscribed_id = listing.subscribed_id

		# Existence - by id.
		ls = self.listed_subscriber.get(listing.subscribed_id, None)
		if ls is not None:
			if ls[0].home_address == listing.home_address:
				self.trace(f'Subscribe ignored "{search}" (refresh)')
			else:
				self.warning(f'Subscribe ignored "{search}" (same id, multiple addresses)')
			return False

		# and search. Build out required structure.
		sr = self.subscribed_search.get(search, None)
		if sr is None:
			try:
				# Keep a single pre-compiled search machine.
				r = re.compile(listing.search)
			except re.error as e:
				t = str(e)
				self.warning(f'Cannot subscribe to "{search}" ({t})')
				return False
			s = {}
			sr = [s, r]
			self.subscribed_search[search] = sr
		else:
			s = sr[0]

		# Existence of subscriber.
		if subscribed_id in s:
			self.warning(f'Subscribe ignored "{search}" (id already registered)')
			return False

		self.console(f'Subscribed "{search}"[{self.directory_scope}]')

		ls = (listing, subscribe)
		s[subscribed_id] = ls
		self.listed_subscriber[subscribed_id] = ls

		if origin is not None:
			origin.add(listing.subscribed_id)
		return True

	def find_subscribers(self, published: Published):
		# Turn a search into a flat list of matching subscribers.
		for k, v in self.subscribed_search.items():
			m = v[1].fullmatch(published.name)
			if m:
				for s in v[0].values():
					yield s[0]

	def find_publishers(self, subscribed: Subscribed):
		# Turn a search into a flat list of matching publishers.
		sr = self.subscribed_search.get(subscribed.search, None)
		if sr is None:
			return
		machine = sr[1]
		for k, lp in self.published_name.items():
			m = machine.fullmatch(k)
			if m:
				yield lp[0]

	def create_route(self, subscriber: Subscribed, publisher: Published):
		self.console(f'Route', name=publisher.name, scope=self.directory_scope, encrypted=publisher.encrypted)
		# There is a match at this scope between given sub and pub.
		# Create the appropriate route object and record its existence.
		# Communication with relevant directories is up to the route.

		route_id = uuid.uuid4()
		if self.directory_scope in (ScopeOfDirectory.LAN, ScopeOfDirectory.HOST, ScopeOfDirectory.GROUP):
			r = self.create(RouteOverConnect,
				route_id=route_id, scope=self.directory_scope, encrypted=publisher.encrypted,
				subscriber=subscriber, publisher=publisher)

		elif self.directory_scope == ScopeOfDirectory.PROCESS:
			s = self.listed_subscriber[subscriber.subscribed_id][1]
			p = self.listed_publisher[publisher.published_id][1]

			r = self.create(RouteInProcess, route_id=route_id,
				subscriber=subscriber, publisher=publisher,
				subscribe_to=s, publish_as=p)

		else:
			self.warning(f'Cannot route "{publisher.name}" at [{self.directory_scope}]')
			return

		# When this publisher is cleared, nudge this route.
		pr = self.routed_publish.get(publisher.published_id, None)
		if pr is None:
			pr = [publisher, set()]
			self.routed_publish[publisher.published_id] = pr
		pr[1].add(r)

		# When this subscriber is cleared, nudge this route.
		sr = self.routed_subscribe.get(subscriber.subscribed_id, None)
		if sr is None:
			sr = [subscriber, set()]
			self.routed_subscribe[subscriber.subscribed_id] = sr
		sr[1].add(r)

		# When the route terminates, clear out the links.
		def clear(self, value, args):
			pr = self.routed_publish.get(args.published_id)
			if pr is not None:
				self.console(f'Cleared route "{pr[0].name}"[{self.directory_scope}] ({args.published_id})')

				pr[1].discard(args.route)
				if len(pr[1]) == 0:
					self.routed_publish.pop(args.published_id, None)
			else:
				self.console(f'Route not cleared "{args.published_id}" [{self.directory_scope}]')

			sr = self.routed_subscribe.get(args.subscribed_id)
			if sr is not None:
				sr[1].discard(args.route)
				if len(sr[1]) == 0:
					self.routed_subscribe.pop(args.subscribed_id, None)

		self.on_return(r, clear, subscribed_id=subscriber.subscribed_id, published_id=publisher.published_id, route=r)

	def clear_listings(self, subscribers, publishers):
		stop = Stop()
		# Remove the listed subscriber ids from this directory.
		# Routing, listings and searching maps need to be popped.
		for s in subscribers:
			# Terminate all routes involving this subscriber.
			routed_subscribe = self.routed_subscribe.get(s, None)
			if routed_subscribe:
				for a in routed_subscribe[1]:
					# Entry cleared by termination of route.
					self.send(stop, a)

			# Remove from the listings.
			listed_subscribe = self.listed_subscriber.pop(s, None)
			if listed_subscribe is None:
				continue
			search = listed_subscribe[0].search

			# Remove from the matching machinery.
			subscribed = self.subscribed_search.get(search, None)
			if subscribed is None:
				continue
			a, m = subscribed
			a.pop(s, None)

			self.console(f'Cleared subscribed "{search}"[{self.directory_scope}]')

			# If this is the home directory, remove from uniqueness check.
			if listed_subscribe[1] is not None:
				unique_subscribe = (listed_subscribe[1].search, listed_subscribe[1].subscriber_address)
				self.unique_subscribe.pop(unique_subscribe, None)

		# Remove the listed publisher ids from this directory.
		for p in publishers:
			# Terminate all routes involving this publisher.
			routed_publish = self.routed_publish.get(p, None)
			if routed_publish:
				for a in routed_publish[1]:
					# Entry cleared by termination of route.
					self.send(stop, a)

			# Remove from the listings.
			listed_publish = self.listed_publisher.pop(p, None)
			if listed_publish is None:
				continue
			name = listed_publish[0].name

			# Remove from the matching machinery.
			self.published_name.pop(name, None)

			self.console(f'Cleared published "{name}"[{self.directory_scope}]')

			# If this is the home directory, remove from uniqueness check.
			if listed_publish[1] is not None:
				unique_publish = listed_publish[1].name
				self.unique_publish.pop(unique_publish, None)

	def open_route(self, route):
		# Callback on loss of ConnectToPeer.
		def clear_ipp(self, value, args):
			self.console(f'Clearing peer connection {args.ipp}')
			self.peer_connect.pop(args.ipp, None)

		def response_to_request(self, value, args):
			r = args.request
			sr = self.subscriber_routing.get(r.subscribed_id, None)
			if sr is None:
				return
			if isinstance(value, Faulted):
				self.console(f'Cannot establish session for {r.subscribed_id} ({value})')
				return
			arc = sr.get(r.name, None)
			if arc is None:
				return
			arc[2] = value

		# Initiate the given route. This is on a per-type basis.
		# Should be a virtual method.
		self.trace(f'Opening route "{route.name}"[{route.scope}]')

		if isinstance(route, RouteOverLoop):
			# Comms is over a standard message connection between this process
			# and the process at the given network address.
			ls = self.listed_subscriber.get(route.subscribed_id, None)
			c = self.peer_connect.get(route.ipp, None)
			if c is None:
				c = self.create(ConnectToPeer, route.ipp, route.encrypted)
				self.on_return(c, clear_ipp, ipp=route.ipp)
				self.peer_connect[route.ipp] = c
			address = ls[1].subscriber_address

			# Initiate loop over this connection for subscriber/name relation.
			r = RequestLoop(name=route.name, scope=route.scope, route_id=route.route_id,
				subscribed_id=route.subscribed_id, subscriber_address=address,
				published_id=route.published_id,
			)
			a = self.create(GetResponse, r, c)
			self.on_return(a, response_to_request, request=r)

		elif isinstance(route, RouteToAddress):
			# Comms is between objects within this process, or this process and a library.
			a = Available(name=route.name, scope=route.scope, route_id=route.route_id,
				subscribed_id=route.subscribed_id, published_id=route.published_id,
				publisher_address=route.publisher_address, opened_at=route.opened_at)

			d = Delivered(name=route.name, scope=route.scope, route_id=route.route_id,
				subscribed_id=route.subscribed_id, published_id=route.published_id,
				subscriber_address=route.subscriber_address, opened_at=route.opened_at)

			self.forward(a, route.subscriber_address, route.publisher_address)
			self.forward(d, route.publisher_address, route.subscriber_address)
		else:
			self.warning(f'Routing by {type(route)} not implemented')

	def drop_route(self, ls, route):
		self.trace(f'Dropping route "{route.name}"[{route.scope}]')

		if isinstance(route, RouteOverLoop):
			# Comms is over actual transport.
			c = self.peer_connect.get(route.ipp, None)
			if c is None:
				return

			def dropped(self, value, args):
				if not isinstance(value, LoopDropped):
					self.warning(f'Unexpected response to drop of loop ({value})')
				route = args.route
				try:
					routing = self.subscriber_routing[route.subscribed_id][route.name]
				except (KeyError, IndexError):
					return READY

				if routing[0] is None:
					pass
				elif routing[0].route_id == route.route_id:
					pass
				else:
					return READY		# Routed by some other means.

				shortest = shortest_route(routing[1])	# Get the best.
				if shortest is None:					# Nothing there.
					return READY
				routing[0] = shortest
				self.open_route(shortest)

			d = DropLoop(name=route.name, scope=route.scope, route_id=route.route_id,
				subscribed_id=route.subscribed_id, published_id=route.published_id,
				subscriber_address=ls[1].subscriber_address)

			a = self.create(GetResponse, d, c, seconds=COMPLETE_A_LOOP)
			self.on_return(a, dropped, route=route)

		elif isinstance(route, RouteToAddress):
			s = Dropped(name=route.name, scope=route.scope, route_id=route.route_id,
				subscribed_id=route.subscribed_id, published_id=route.published_id,
				remote_address=route.publisher_address, opened_at=route.opened_at)

			p = Dropped(name=route.name, scope=route.scope, route_id=route.route_id,
				subscribed_id=route.subscribed_id, published_id=route.published_id,
				remote_address=route.subscriber_address, opened_at=route.opened_at)

			self.forward(s, route.subscriber_address, route.publisher_address)
			self.forward(p, route.publisher_address, route.subscriber_address)

	def clear_route(self, ls, route):
		self.trace(f'Clearing route "{route.name}"[{route.scope}]')

		if isinstance(route, RouteOverLoop):
			# Transport-based pub-sub is not cleared on loss of the
			# control channel.
			return False

		elif isinstance(route, RouteToAddress):
			s = Dropped(name=route.name, scope=route.scope, route_id=route.route_id,
				subscribed_id=route.subscribed_id, published_id=route.published_id,
				remote_address=route.publisher_address, opened_at=route.opened_at)

			p = Dropped(name=route.name, scope=route.scope, route_id=route.route_id,
				subscribed_id=route.subscribed_id, published_id=route.published_id,
				remote_address=route.subscriber_address, opened_at=route.opened_at)

			self.forward(s, route.subscriber_address, route.publisher_address)
			self.forward(p, route.publisher_address, route.subscriber_address)

			return True
		return False

	def send_up(self, listing):
		if isinstance(self.connected, Connected):
			if listing.scope.value < self.directory_scope.value:
				self.send(listing, self.connected.proxy_address)

	def push_up(self):
		published = [v[0] for k, v in self.published_name.items() if v[0].scope.value < self.directory_scope.value]
		subscribed = []
		for k, v in self.subscribed_search.items():
			for t in v[0].values():
				if t[0].scope.value < self.directory_scope.value:
					subscribed.append(t[0])
		if published or subscribed:
			self.send(PublishedDirectory(published, subscribed), self.connected.proxy_address)

	def get_directory(self, client_address):
		argv = sys.argv
		if argv[0].rfind('python') == -1:
			name = argv[0]
		else:
			name = argv[1]
		executable = os.path.basename(name)

		published = [v[0] for k, v in self.listed_publisher.items()]
		subscribed = [v[0] for k, v in self.listed_subscriber.items()]

		directory_route = {}
		for k, v in self.routed_publish.items():
			pub, routed = v
			for r in routed:
				directory_route[r] = DirectoryRoute(name=pub.name, published_id=pub.published_id)

		for k, v in self.routed_subscribe.items():
			sub, routed = v
			for r in routed:
				dr = directory_route.get(r, None)
				if dr is None:
					continue
				dr.subscribed_id = sub.subscribed_id

		routed = [v for k, v in directory_route.items() if v.subscribed_id and v.published_id]

		peer = {}
		for subscribed_id, v in self.subscriber_routing.items():
			ls = self.listed_subscriber.get(subscribed_id, None)
			if ls is None:
				continue
			s = [PeerSession(name=name, route=t3[0], status=t3[2]) for name, t3 in v.items()]
			c = DirectoryPeer(subscribed_id=subscribed_id, search=ls[0].search, session=s)
			peer[subscribed_id] = c

		primary_ip = get_local_ip()

		sub_directory = {}
		if isinstance(self.listening, Listening):
			accept_directories_at = self.listening.listening_ipp
			for k, v in self.accepted.items():
				ta, sub, pub = v
				sub_directory[ta.opened_ipp] = ta.proxy_address
		else:
			accept_directories_at = self.accept_directories_at

		directory = DirectoryListing(unique_id=self.unique_id, executable=executable,
			directory_address=self.object_address, scope=self.directory_scope,
			primary_ip=primary_ip,
			connect_to_directory=self.connect_to_directory, accept_directories_at=accept_directories_at,
			published=published, subscribed=subscribed,
			routed=routed, peer=peer, sub_directory=sub_directory)

		self.send(directory, client_address)

#
#
def ObjectDirectory_INITIAL_Start(self, message):
	self.calculate_reconnect(self.connect_to_directory.host)
	if self.connect_to_directory.host is not None:
		connect(self, self.connect_to_directory,
			keep_alive=self.keep_alive(),
			encrypted=self.encrypted_directory())
	if self.accept_directories_at.host is not None:
		listen(self, self.accept_directories_at,
			encrypted=self.encrypted_directory())
	return READY

#
def ObjectDirectory_READY_Listening(self, message):
	self.listening = message
	for p in self.pending_enquiry:
		self.send(message.listening_ipp, p)
	self.pending_enquiry = set()
	return READY

def ObjectDirectory_READY_NotListening(self, message):
	self.listening = message
	# Schedule a retry.
	return READY

def ObjectDirectory_READY_Connected(self, message):
	self.connected = message
	self.push_up()
	if self.directory_opened:
		self.send(DirectoryOpened(), self.directory_opened)
		self.directory_opened = None
	return READY

def ObjectDirectory_READY_NotConnected(self, message):
	if self.connect_to_directory.host:
		self.connected = message
		self.start(T1, self.reconnect_delay)
		if self.directory_opened:
			self.send(DirectoryOpened(), self.directory_opened)
			self.directory_opened = None
	return READY

def ObjectDirectory_READY_T1(self, message):
	if self.connect_to_directory.host is not None:
		connect(self, self.connect_to_directory,
			keep_alive=self.keep_alive(),
			encrypted=self.encrypted_directory())
	return READY

def ObjectDirectory_READY_Accepted(self, message):
	self.accepted[self.return_address[-1]] = [message, set(), set()]	# Accepted, subs, pubs.
	return READY

def ObjectDirectory_READY_Closed(self, message):
	if isinstance(self.connected, Connected):
		if self.return_address == self.connected.proxy_address:
			self.connected = message
			self.start(T1, self.reconnect_delay)
			return READY

	p = self.accepted.pop(self.return_address[-1], None)
	if p is None:
		return READY
	self.clear_listings(p[1], p[2])
	if isinstance(self.connected, Connected) and (p[1] or p[2]):
		self.send(ClearListings(p[1], p[2]), self.connected.proxy_address)
	return READY

def ObjectDirectory_READY_Enquiry(self, message):
	if self.accept_directories_at.host is None:
		self.accept_directories_at = DIRECTORY_AT_EPHEMERAL
		listen(self, self.accept_directories_at,
			encrypted=self.encrypted_directory())
		self.pending_enquiry.add(self.return_address)
		return READY

	if not isinstance(self.listening, Listening):
		self.pending_enquiry.add(self.return_address)
	else:
		self.reply(self.listening.listening_ipp)

	return READY

def ObjectDirectory_READY_PublishAs(self, message):
	name = message.name
	scope = message.scope
	encrypted = message.encrypted

	unique_publish = name
	r = self.unique_publish.get(unique_publish, None)
	if r is not None:
		self.reply(NotPublished(name=name, scope=scope, note=f'already published'))
		return READY

	self.auto_connect(message)

	published_id = uuid.uuid4()
	listing = Published(name=name, scope=scope, encrypted=encrypted,
		published_id=published_id, home_address=self.object_address)

	if not self.add_publisher(listing, None, message):
		return READY
	self.unique_publish[unique_publish] = published_id
	if self.directory_scope.value < ScopeOfDirectory.LIBRARY.value:
		self.send(listing, message.publisher_address)

	for s in self.find_subscribers(listing):
		self.create_route(s, listing)
	self.send_up(listing)

	return READY

def ObjectDirectory_READY_HostPort(self, message):
	d = self.progress()
	if d is None:
		self.warning(f'cannot complete ListenForPeer (no message record)')
		return READY
	listing, publish = d
	listing.listening_ipp = message		# Update with live network address.

	self.console(f'Published[{self.directory_scope}]', name=listing.name, listening=message)

	self.published_name[listing.name] = (listing, publish)
	self.listed_publisher[listing.published_id] = (listing, publish)

	unique_publish = publish.name
	self.unique_publish[unique_publish] = listing.published_id
	self.send(listing, publish.publisher_address)

	for s in self.find_subscribers(listing):
		self.create_route(s, listing)
	self.send_up(listing)

	return READY

def ObjectDirectory_READY_SubscribeTo(self, message):
	search = message.search
	scope = message.scope

	unique_subscribe = (search, self.return_address)
	r = self.unique_subscribe.get(unique_subscribe, None)
	if r is not None:
		self.reply(NotSubscribed(search=search, scope=scope, note=f'already subscribed'))
		return READY

	self.auto_connect(message)

	subscribed_id = uuid.uuid4()
	listing = Subscribed(search=search, scope=scope, subscribed_id=subscribed_id, home_address=self.object_address)
	if not self.add_subscriber(listing, None, message):
		self.reply(NotSubscribed(search=search, scope=scope, note=f'duplicates/search expression'))
		return READY
	self.unique_subscribe[unique_subscribe] = subscribed_id
	self.send(listing, message.subscriber_address)

	for p in self.find_publishers(listing):
		self.create_route(listing, p)
	self.send_up(listing)

	return READY

def ObjectDirectory_READY_Published(self, message):
	if message.scope.value > self.directory_scope.value:
		return READY

	a, sub, pub = self.accepted[self.return_address[-1]]
	if not self.add_publisher(message, pub, None):
		self.send(Advisory(name=message.name, scope=self.directory_scope, published_id=message.published_id), message.home_address)
		return READY
	self.auto_connect(message)

	for s in self.find_subscribers(message):
		self.create_route(s, message)
	self.send_up(message)

	return READY

def ObjectDirectory_READY_Advisory(self, message):
	self.warning(f'Cannot publish "{message.name}" at [{message.scope}]')
	return READY

def ObjectDirectory_READY_Subscribed(self, message):
	if message.scope.value > self.directory_scope.value:
		return READY

	a, sub, pub = self.accepted[self.return_address[-1]]
	if not self.add_subscriber(message, sub, None):
		return READY
	self.auto_connect(message)

	for p in self.find_publishers(message):
		self.create_route(message, p)
	self.send_up(message)
	return READY

def ObjectDirectory_READY_PublishedDirectory(self, message):
	a, sub, pub = self.accepted[self.return_address[-1]]
	highest = None
	for p in message.published:
		if p.scope.value > self.directory_scope.value:
			continue
		if not self.add_publisher(p, pub, None):
			continue
		if highest is None or p.scope.value < highest.scope.value:
			highest = p
		for s in self.find_subscribers(p):
			self.create_route(s, p)

	highest is not None and self.auto_connect(highest)

	for s in message.subscribed:
		if s.scope.value > self.directory_scope.value:
			continue
		if not self.add_subscriber(s, sub, None):
			continue
		if highest is None or s.scope.value < highest.scope.value:
			highest = s
		for p in self.find_publishers(s):
			self.create_route(s, p)

	highest is not None and self.auto_connect(highest)

	if isinstance(self.connected, Connected):
		self.push_up()
	return READY

def ObjectDirectory_READY_ClearListings(self, message):
	self.clear_listings(message.subscribers, message.publishers)
	if isinstance(self.connected, Connected):
		self.send(message, self.connected.proxy_address)
	return READY

def ObjectDirectory_READY_ClearPublished(self, message):
	subscribers = set()
	publishers = set([message.published_id])
	self.clear_listings(subscribers, publishers)
	self.reply(PublishedCleared(name=message.name, scope=message.scope, published_id=message.published_id, note=message.note))
	if isinstance(self.connected, Connected):
		self.send(ClearListings(subscribers, publishers), self.connected.proxy_address)
	return READY

def ObjectDirectory_READY_ClearSubscribed(self, message):
	subscribers = set([message.subscribed_id])
	publishers = set()
	self.clear_listings(subscribers, publishers)
	self.reply(SubscribedCleared(search=message.search, scope=message.scope, subscribed_id=message.subscribed_id, note=message.note))
	if isinstance(self.connected, Connected):
		self.send(ClearListings(subscribers, publishers), self.connected.proxy_address)
	return READY

def ObjectDirectory_READY_ClearSubscriberRoute(self, message):
	subscribed_id = message.subscribed_id
	listing = self.listed_subscriber.get(subscribed_id, None)
	if listing is None:
		self.warning('No such subscription')
		return READY

	# Find the routing table.
	subscriber = self.subscriber_routing.get(subscribed_id, None)
	if subscriber is None:
		return READY

	routing = subscriber.get(message.name, None)
	if routing is None:
		return READY

	# Delete the route from the available routes.
	d = delete_route(message.route_id, routing[1])
	if not d:
		self.warning('Unknown route')
		return READY

	if routing[0] and routing[0].route_id == message.route_id:
		if self.clear_route(listing, routing[0]):
			routing[0] = None

	# Clear out empty vessels.
	if len(routing[1]) == 0:
		subscriber.pop(message.name, None)

	if len(subscriber) == 0:
		self.subscriber_routing.pop(subscribed_id, None)
	return READY

def ObjectDirectory_READY_ClearPublisherRoute(self, message):
	p = self.listed_publisher.get(message.published_id, None)
	if p is None:
		return READY
	# So far there is no action required on the
	# publisher side.
	return READY

def ObjectDirectory_READY_OpenLibrary(self, message):
	if message.published_id:
		p = self.listed_publisher.get(message.published_id, None)
		if p[1]:
			self.reply(p[1])
	elif message.subscribed_id:
		s = self.listed_subscriber.get(message.subscribed_id, None)
		if s[1]:
			self.reply(s[1])
	return READY

def ObjectDirectory_READY_SubscriberRoute(self, message):
	subscribed_id = message.subscribed_id

	# Add the route to the routing[subscriber][name] table.
	# Evaluate the (changed) routing options.
	# If no current loop, initiate the best option.
	ls = self.listed_subscriber.get(subscribed_id, None)
	if ls is None or ls[1] is None:
		self.warning('No such subscription or not the home directory')
		return READY

	# Routing per listing
	subscriber = self.subscriber_routing.get(subscribed_id, None)
	if subscriber is None:
		subscriber = {}
		self.subscriber_routing[subscribed_id] = subscriber

	# Per matched name
	routing = subscriber.get(message.name, None)
	if routing is None:
		routing = [None, [], ToBeConfirmed('no active route or loop not closed')]
		subscriber[message.name] = routing

	# Final checks.
	i = scope_route(message, routing[1])
	if i is not None:
		if isinstance(message, RouteOverLoop):
			if equal_ipp(message.ipp, routing[1][i].ipp):
				self.trace(f'Replacement route at [{message.scope}]')
				routing[1][i] = message
				# Cant do this - would invalidate session notifications.
				#if routing[0].scope == message.scope:
				#	routing[0] = message
				return READY
		self.trace(f'Duplicate route at [{message.scope}]')
		return READY

	# Evaluate what affect the addition of this route will
	# have on this subscriber-to-name connection effort.
	shortest = add_route(message, routing[1])

	if routing[0] is not None:	# Route is active.
		if shortest.scope.value > routing[0].scope.value:
			self.trace(f'Upgrading "{routing[0].name}"[{routing[0].scope}] to inner [{shortest.scope}]')
			self.drop_route(ls, routing[0])
		else:
			self.trace(f'Added outer route "{message.name}"[{message.scope}]')
		return READY

	routing[0] = shortest
	self.open_route(shortest)
	return READY

def ObjectDirectory_READY_LoopDropped(self, message):
	try:
		routing = self.subscriber_routing[message.subscribed_id][message.name]
	except (KeyError, IndexError):
		return READY

	if routing[0] and routing[0].route_id == message.route_id:
		routing[0] = None
		shortest = shortest_route(routing[1], excluding=message.route_id)
		if shortest is None:
			return READY
		routing[0] = shortest
		self.open_route(shortest)
	return READY

def ObjectDirectory_READY_Returned(self, message):
	d = self.debrief()
	if isinstance(d, OnReturned):
		d(self, message)
	return READY

def ObjectDirectory_READY_Stop(self, message):
	self.complete()

import json
def ObjectDirectory_READY_Incognito(self, message):
	s = json.dumps(message.decoded_word)
	self.console(type_name=message.type_name, word=s)
	return READY

def ObjectDirectory_READY_OpenDirectory(self, message):
	self.directory_opened = self.return_address

	if self.connected:
		self.reply(DirectoryOpened())
		return READY

	if self.connect_to_directory.host is None:
		pa = PublishAs(name='open-directory', scope=message.scope)
		self.auto_connect(pa)

	return READY

def ObjectDirectory_READY_ListDirectory(self, message):
	if isinstance(self.connected, Connected):
		self.forward(message, self.connected.proxy_address, self.return_address)
		return READY

	self.get_directory(self.return_address)
	return READY

def ObjectDirectory_READY_GetDirectory(self, message):
	self.get_directory(self.return_address)
	return READY

def ObjectDirectory_READY_Ping(self, message):
	self.reply(message)
	return READY

OBJECT_DIRECTORY_DISPATCH = {
	INITIAL: (
		(Start,),
		()
	),
	READY: (
		(Listening, NotListening,
		Connected, NotConnected,
		T1,
		Accepted, Closed,
		Enquiry,
		PublishAs, SubscribeTo, HostPort,
		Published, Subscribed,
		Advisory,
		PublishedDirectory,
		ClearListings,
		ClearPublished, ClearSubscribed,
		ClearSubscriberRoute, ClearPublisherRoute,
		LoopDropped,
		OpenLibrary,
		SubscriberRoute,
		Returned,
		Stop,
		Incognito,
		OpenDirectory, ListDirectory, GetDirectory,
		Ping,),
		()
	),
}

bind(ObjectDirectory, OBJECT_DIRECTORY_DISPATCH)



# test_directory.py
# Demonstration of builtin directory behaviour.
# Configuration happens on the command line, e.g. --connect-to-directory=<HostPort>
# Not available as sticky settings.
import kipjak as kj

def directory(self):
	'''Hold the foreground while directory operates in the background. Return notification.'''

	while True:
		m = self.input()
		if isinstance(m, kj.Stop):
			return kj.Aborted()

# Register with runtime.
kj.bind(directory)

# Process entry-point.
if __name__ == '__main__':
	kj.create(directory)



# test_storage.py
import kipjak as kj

#
#
def delta(self, left_path: str=None, right_path: str=None) -> list[kj.Any]:
	'''.'''
	if left_path is None or right_path is None:
		return kj.Faulted('no storage specified.')

	lp = kj.storage_manifest(left_path)
	rp = kj.storage_manifest(right_path)

	m = [d for d in kj.storage_delta(lp[0], rp[0])]

	return m

kj.bind(delta)	# Register with the framework.



if __name__ == '__main__':	# Process entry-point.
	kj.create(delta)



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Running a function within its own thread.

"""
__docformat__ = 'restructuredtext'

import threading
import queue

from .virtual_memory import *
from .convert_memory import *
from .virtual_runtime import *
from .point_runtime import *
from .object_space import *

__all__ = [
	'start_a_thread',
	'running_in_thread',
]

#
#
def start_a_thread(queue, routine, args, kw_args):
	"""
	Start a platfrom thread that runs the given custom routine on the
	given queue. No return.

	This is the first part of arranging a function to run in its own
	dedicated thread, with its own unique id and a `self` parameter it
	can use to interact with the rest of the world. See
	`running_in_thread` for the second part.

	It is an error to call this function directly from an application.
	Instead this is a function that is called from ``Point`` methods
	such as ``custom_routine`` and ``object_and_thread``. An
	application that makes direct calls to this function will
	enjoy undefined behaviour.

	Parameters:

	- `queue`: an async object, a Queue-based machine or Channel.
	- `routine`: a plain, old function, the code that will execute
	within the new thread.
	- `args`: tuple of values, the args forwarded on to the routine(...)
	call.
	- `kw_args`: dict of named values, the named args forwarded on to the
	routine(...) call.

	Returns:

	Nothing.
	"""
	queue.thread_function = routine
	queue.assigned_thread = threading.Thread(target=running_in_thread, args=(routine, queue, args, kw_args))
	queue.assigned_thread.daemon = True
	queue.assigned_thread.start()

def running_in_thread(routine, queue, args, kw_args):
	"""
	First code to run inside new thread. Set up some kipjak
	admin information before passing control to the true
	tenant of this thread, i.e. the custom routine.

	Also serves as the wrapper for management of termination mechanisms.

	Parameters:

	- `routine`: a plain, old function, the code to be called.
	- `queue`: an async object, a Queue-based machine or Channel that
	becomes `self`.
	- `args`: tuple of values, the args forwarded on to the routine(...)
	call.
	- `kw_args`: dict of named values, the named args forwarded on to the
	routine(...) call.

	Returns:

	Nothing.
	"""
	address = queue.object_address
	parent = queue.parent_address

	value = None
	# Need to catch all exits from this application-provided
	# code, to help maintain integrity of object map. But also
	# need to allow system exceptions pass through.
	try:
		message = routine(queue, *args, **kw_args)
	# Necessary replication of exceptions in
	# object_dispatch.
	except KeyboardInterrupt:
		s = 'unexpected keyboard interrrupt'
		queue.fault(s)
		message = Faulted('object compromised', s)
	except SystemExit:
		s = 'unexpected system exit'
		queue.fault(s)
		message = Faulted('object compromised', s)
	except Completion as c:
		# From run_object (threads dedicated to machines), object_dispatch (e.g. class
		# threads) and all custom routines.
		message = c.message
	except Exception as e:
		s = str(e)
		s = f'unhandled exception ({s})'
		queue.fault(s)
		message = Faulted('object faulted', s)
	except:
		s = 'unhandled opaque exception'
		queue.fault(s)
		message = Faulted('object faulted', s)

	if queue.__art__.lifecycle:
		queue.log(USER_TAG.DESTROYED, 'Destroyed')

	return_type = routine.__art__.return_type
	if return_type is None:
		pass
	elif isinstance(return_type, Any):
		pass
	elif isinstance(return_type, Portable):
		if not hasattr(message, '__art__'):
			message = (message, return_type)
	else:
		message = Faulted(f'unexpected return type for routine "{routine.__name__}"')

	ending = queue.object_ending
	destroy_an_object(address)
	ending(message, parent, address)



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module defines the machinery that underpins the timer service
available at `Point.start`. There are three major parts:

- ``Tick``, a class deriving from Message, this is the metronome sent to
  countdown machine.
- ``CountdownTimer``, a class, runs in own thread this object accepts
  StartTimer messsages.
- ``timer_circuit``, a function, runs as custom routine, sends regular
  Ticks to CountdownTimer.
"""
__docformat__ = 'restructuredtext'

from time import sleep, monotonic

from .object_space import *
from .virtual_runtime import *
from .point_runtime import *
from .virtual_point import *
from .message_memory import *
from .routine_point import *
from .point_machine import *
from .bind_type import *

__all__ = [
	'Tick',
	'CountdownTimer',
	'timer_circuit',
]

class Tick(object):
	pass

bind_message(Tick,
	message_trail=False, execution_trace=False, copy_before_sending=False)

# Handling of expression of time values
# and conversion to something that this
# module understands.
TICKS_PER_SECOND = 4
TIME_QUANTUM = 100 / TICKS_PER_SECOND
QUANTUM_LESS_1 = TIME_QUANTUM - 1

class CountdownTimer(Threaded, Stateless):
	"""
	An async object running in its own thread. Available to all `Point`
	objects as pt.timer_address.

	This object accepts StartTimer messages, creating a record of each
	active timer as if there was a (sender_address, timer) unique,
	composite key. The record is initialized with the ticks before
	expiry.

	Ticks are received from the timer_circuit and are used to decrement
	the active record. If this decements to less than 1 the record is
	"expired". The timer message is sent and the record is removed.

	Most timer-related logging is necessarily disabled.
	"""
	def __init__(self):
		Threaded.__init__(self)
		Stateless.__init__(self)
		self.running = []

def CountdownTimer_Start(self, message):
	# Silence the "dropped" message.
	pass

def CountdownTimer_Tick(self, message):
	still_running = []
	now = monotonic()
	for c in self.running:
		if now >= c[4]:			# END OF A PERIOD
			who = c[0]			# Who requested it.
			what = c[1]()		# The timer instance.
			send_a_message(what, who, self.object_address)
			if c[3]:
				c[4] = monotonic() + c[2]
				still_running.append(c)
		else:
			still_running.append(c)

	if len(still_running) != len(self.running):
		self.running = still_running

def CountdownTimer_StartTimer(self, message):
	sender  = self.return_address
	timer   = message.timer
	for c in self.running:
		if c[0] == sender and c[1] == timer:	# An existing timer.
			c[2] = message.seconds				# Redefine its details.
			c[3] = message.repeating
			c[4] = monotonic() + c[2]			# Freshen up that expiry.
			return

	if message.seconds < 0.2:
		send_a_message(message.timer(), sender, self.object_address)
		return

	# A brand new timer block;
	# [0] address of requesting party
	# [1] class, e.g. ar.T1
	# [2] seconds of delay
	# [3] true if a repeating timer
	# [4] the moment of expiry.
	c = [
		sender,				# Who wants it.
		timer,				# Class of the timer.
		message.seconds,
		message.repeating,
		monotonic() + message.seconds
	]
	self.running.append(c)

def CountdownTimer_CancelTimer(self, message):
	return_address = self.return_address
	for i, c in enumerate(self.running):
		if c[0] == return_address and c[1] == message.timer:
			# Ok to modify cos immediate
			# return afterwards?
			self.running.pop(i)
			return

def CountdownTimer_Stop(self, message):
	self.complete()

bind_stateless(CountdownTimer,
	(Start, Tick, StartTimer, CancelTimer, Stop),
	lifecycle=False, message_trail=False,
	execution_trace=False, user_logs=USER_LOG.NONE)

# Control the rate of ticks
BETWEEN_TICKS = 1.0 / TICKS_PER_SECOND

def timer_circuit(queue, timer_address):
		t = Tick()
		while not queue.halted:
			send_a_message(t, timer_address, queue.object_address)
			sleep(BETWEEN_TICKS)
		t = Stop()

bind_routine(timer_circuit,
	lifecycle=False, message_trail=False, execution_trace=False, user_logs=USER_LOG.NONE)



# test_main_args.py
import datetime
import uuid
import kipjak as kj

from test_person import *

def main(self, table: dict[str,list[Person]]=None,
	count: int=10, ratio: float=0.5,
	who: Person=None, when: datetime.datetime=None,
	unique_id: uuid.UUID=None):

	# Traditional defaults.
	table = table or dict(recent=[Person('Felicity'), Person('Frederic')])
	who = who or Person('Wilfred')
	when = when or kj.world_now()
	unique_id = unique_id or uuid.uuid4()

	# Name of every person.
	person = []
	for v in table.values():
		for p in v:
			person.append(p.given_name)

	# Log values.
	csv = ','.join(person)
	self.console(table=csv)
	self.console(count=count, ratio=ratio)
	self.console(who=who.given_name, when=when)
	self.console(unique_id=unique_id)

kj.bind(main)

if __name__ == '__main__':
	kj.create(main)



# object_startup_test.py
import datetime
import kipjak as kj

from test_person import *

table_Person_type = kj.def_type(kj.VectorOf(kj.VectorOf(kj.UserDefined(Person))))
table_world_type = kj.def_type(kj.VectorOf(kj.VectorOf(kj.WorldTime())))

def main(self, height: int=4, width: int=4, who: Person=None, when: datetime.datetime=None):
	self.console(f'width: {width}, height: {height}')

	if height < 1 or height > 1000 or width < 1 or width > 1000:
		return kj.Faulted(f'out of bounds')

	if who:
		table = [[who] * width] * height
		return kj.cast_to(table, table_Person_type)

	if when:
		table = [[when] * width] * height
		return kj.cast_to(table, table_world_type)

	return (True, kj.Boolean())

kj.bind(main, return_type=kj.Any())

if __name__ == '__main__':
	kj.create(main)



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Platform processes as objects.

The ProcessObject machine creates and manages a platform process. Arguments
are passed to the new process and a value extracted from standard output is
repurposed as the return value.

If an API is defined for the object, a listen port is provided to the new
process as an argument and the framework arranges for a connection from the
ProcessObject to the object created inside the sub-process (see publish() and
subscribe()).
"""
__docformat__ = 'restructuredtext'

import sys
import os
import signal
import shutil
from subprocess import Popen, PIPE
from collections import deque

from .general_purpose import *
from .virtual_memory import *
from .convert_memory import *
from .message_memory import *
from .convert_signature import *
from .convert_type import *
from .virtual_runtime import *
from .virtual_point import *
from .point_runtime import *
from .routine_point import *
from .virtual_codec import *
from .json_codec import *
from .noop_codec import *
from .point_machine import *
from .object_runtime import *
from .command_line import *
from .command_startup import *
from .object_startup import *
from .home_role import *
from .bind_type import *
from .object_collector import *
from .object_directory import *
from .process_directory import *
from .ip_networking import *
from .get_response import *

__all__ = [
	'ProcessObject',
	'Utility'
]

PO = Gas(collector=None)

# Managed creation of processes.
def create_processes(root):
	PO.collector = root.create(ObjectCollector)

def stop_processes(root):
	root.send(Stop(), PO.collector)
	root.select()

AddOn(create_processes, stop_processes)

# A thread dedicated to the blocking task of
# waiting for termination of a process.
class CodePage(object):
	def __init__(self, code: int=None, page: str=None):
		self.code = code
		self.page = page

bind(CodePage)

def wait(self, p, piping):
	if piping:
		out, err = p.communicate()
		return CodePage(p.returncode, out)
	else:
		code = p.wait()
		return CodePage(code, None)

bind(wait)

# The ProcessObject class
# A platform process that accepts a set of fully-typed arguments
# and returns a fully-typed result.
class INITIAL: pass
class PENDING: pass
class EXECUTING: pass
class CLEARING: pass
class SPOOLING: pass

# Control over the family of processes that can result from the initiation
# of a single one. This was quite difficult to get right and requires the
# proper use of a ProcessObject object (paramters origin and debug) to preserve
# the correct behaviour from one component to the next.

class ProcessObject(Point, StateMachine):
	"""
	An async proxy object that starts and manages a platform sub-process.

	:param object_or_name: object to run as a process or the executable file name
	:type object_or_name: :ref:`object type<kj-object-type>` or str
	:param args: positional arguments to be passed to the process
	:param origin: processing context
	:param home_path: location of a composite process
	:param role_name: name within the composite process
	:param top_role: enable top-level behaviour
	:param object_api: enable private library behaviour for named process (str)
	:param extra_types: force loading of library types
	:param remainder_args: forward unknown command-line args to process
	:param settings: named arguments to be encoded for the process
	"""
	def __init__(self, object_or_name, *args, origin: ProcessOrigin=None,
			home_path: str=None, role_name: str=None, top_role: bool=False,
			object_api: list=None, extra_types: list=None,
			remainder_args: list=None, **settings):
		Point.__init__(self)
		StateMachine.__init__(self, INITIAL)
		self.args = args
		self.object_or_name = object_or_name
		self.origin = origin
		self.home_path = home_path
		self.role_name = role_name
		self.top_role = top_role
		self.object_api = object_api
		self.extra_types = extra_types
		self.remainder_args = remainder_args
		self.settings = settings

		self.module_path = None
		self.script_path = None
		self.origin_path = None
		self.api = None
		self.listening_for_directories = None
		self.p = None

		self.published = None
		self.subscribed = None
		self.queue = deque()

	def load_image(self, rt):
		# Start the new process.
		# Derive the home/role context using the globals defined for
		# this process, plus the values describing the new process.

		if self.origin:
			origin = self.origin
		else:
			origin = CL.origin
			if origin in (ProcessOrigin.RUN, ProcessOrigin.RUN_CHILD):
				origin = ProcessOrigin.RUN_CHILD
			elif origin in (ProcessOrigin.START, ProcessOrigin.START_CHILD):
				origin = ProcessOrigin.START_CHILD

		self.role_name = self.role_name or breakpath(self.module_path)[1]
		self.home_path = self.home_path or CL.home_path
		if CL.role_name and not self.top_role:
			self.role_name = f'{CL.role_name}.{self.role_name}'

		s = os.path.splitext(self.module_path)
		dot_py = s[1] and s[1] == '.py'

		# Build the command line.
		command = []
		if dot_py:
			interpreter = sys.executable
			command.append(interpreter)
		command.append(self.module_path)

		if self.args:
			for a in self.args:
				command.append(str(a))

		if self.remainder_args:
			command.extend(remaining_arguments(self.remainder_args))

		try:
			c = CodecNoop()
			if self.listening_for_directories:
				e = self.settings.get('encrypted_process', None)
				encrypted_process = e == True

				v = encode_argument(c, self.listening_for_directories, UserDefined(HostPort))
				b = encode_argument(c, encrypted_process, Boolean())
				command.append(f'--directory-scope=LIBRARY')
				command.append(f'--connect-to-directory={v}')
				command.append(f'--encrypted-process={b}')

				subscribe(self, self.role_name, scope=ScopeOfDirectory.PROCESS)

			# Generate the proper argument strings.
			if rt:
				schema = rt.schema | CommandLine.__art__.schema
			else:
				schema = CommandLine.__art__.schema

			for k, v in self.settings.items():
				if v is None:
					continue
				name = k
				e = schema.get(k, None)
				if e:
					k = k.replace('_', '-')
					v = encode_argument(c, v, e)
				else:
					v = str(v)
				command.append(f'--{k}={v}')
		except CodecError as e:
			e = str(e)
			s = e.replace('cannot encode', f'cannot encode value for argument "{name}", {self.module_path}')
			self.complete(Faulted(s))

		if origin:
			command.append(f'--origin={origin.name}')

		command.append(f'--child-process')
		command.append(f'--full-output')

		if self.home_path:
			command.append(f'--home-path={self.home_path}')
		command.append(f'--role-name={self.role_name}')

		if CL.debug_level is not None and origin not in (ProcessOrigin.START, ProcessOrigin.START_CHILD):
			command.append(f'--debug-level={CL.debug_level.name}')

		if CL.keep_logs:
			command.append(f'--keep-logs')

		if self.script_path or self.origin_path:
			environ = os.environ.copy()
			existing_path = environ.get('PYTHONPATH', None)

			path = []
			if self.script_path:
				path.append(self.script_path)
			if self.origin_path:
				path.append(self.origin_path)
			if existing_path:
				path.append(existing_path)

			python_path = ':'.join(path)
			environ['PYTHONPATH'] = python_path
		else:
			environ = None

		c = ' '.join(command)
		self.console(c)
		try:
			start_new_session = CL.child_process
			self.p = Popen(command,
				#start_new_session=start_new_session,
				stdin=None, stdout=PIPE, stderr=sys.stderr,
				text=True, encoding='utf-8', errors='strict',
				env=environ)
			#**self.kw)
		except OSError as e:
			s = f'cannot start process "{self.module_path}" ({e})'
			self.complete(Faulted(s))

		self.log(USER_TAG.STARTED, f'Started process ({self.p.pid})')
		self.create(wait, self.p, True)

		# Good to go. Next event should be Returned.
		self.send(AddObject(self.object_address), PO.collector)

def find_module(name_py):
	if CL.home_path and CL.role_name:
		# Deployed script.
		candidate = os.path.join(CL.home_path, 'script', name_py)
		if os.path.isfile(candidate):
			return candidate

		# Common ancestry.
		home_role = os.path.join(CL.home_path, CL.role_name)
		resource_path = os.path.join(CL.home_path, 'resource')
		role = open_role(home_role, resource_path)
		if role:
			executable_file = role.executable_file()
			split = os.path.split(executable_file)
			if split[0]:
				candidate = os.path.join(split[0], name_py)
				if os.path.isfile(candidate):
					return candidate

	cwd = os.getcwd()
	candidate = os.path.join(cwd, name_py)
	if os.path.isfile(candidate):
		return candidate

	return None

def find_role(executable_file, home_path):
	script_path = os.path.join(home_path, 'script')

	split = os.path.split(executable_file)
	candidate = os.path.join(script_path, split[1])
	if os.path.isfile(candidate):
		return candidate

	if os.path.isfile(executable_file):
		return executable_file

	return None

def ProcessObject_INITIAL_Start(self, message):
	# Either its a module name, e.g. "test_directory", or its a function/machine
	# to be executed.
	rt = None
	if isinstance(self.object_or_name, str):
		s = os.path.splitext(self.object_or_name)
		if not s[1]:
			self.module_path = shutil.which(s[0])
		elif s[1] == '.py':
			self.module_path = find_module(self.object_or_name)
		else:
			self.complete(Faulted(f'cannot execute {self.object_or_name} (unknown extension?)'))

		if self.module_path is None:
			self.complete(Faulted(f'cannot execute {self.object_or_name} (not found)'))

		if self.object_api is not None and len(self.object_api) > 0:
			self.api = self.object_api
			self.send(Enquiry(), PD.directory)
			return PENDING

	elif isinstance(self.object_or_name, HomeRole):
		executable_file = self.object_or_name.executable_file()
		if not self.home_path:
			self.complete(Faulted(f'cannot execute {executable_file} (no home_path)'))

		self.module_path = find_role(executable_file, self.home_path)
		if self.module_path is None:
			self.complete(Faulted(f'cannot execute {executable_file} (no role script)'))

		split = os.path.split(executable_file)
		origin_path = split[0]
		script_path = os.path.join(self.home_path, 'script')
		if os.path.isdir(origin_path):
			self.origin_path = origin_path
		if os.path.isdir(script_path):
			self.script_path = script_path

	else:
		rt = getattr(self.object_or_name, '__art__', None)
		if rt is None:
			self.complete(Faulted(f'cannot execute {self.object_or_name} (not registered)'))

		imported_module = self.object_or_name.__module__
		module = sys.modules[imported_module]
		self.module_path = module.__file__

		if rt.entry_point is not None and len(rt.entry_point) > 0:
			self.entry_point = rt.entry_point
			self.send(Enquiry(), PD.directory)
			return PENDING

	self.load_image(rt)
	return EXECUTING

def ProcessObject_PENDING_Unknown(self, message):
	message = cast_to(message, self.received_type)
	q = (message, self.return_address)
	self.queue.append(q)
	return PENDING

def ProcessObject_PENDING_HostPort(self, message):
	self.listening_for_directories = message
	rt = getattr(self.object_or_name, '__art__', None)
	self.load_image(rt)
	return EXECUTING

def ProcessObject_EXECUTING_Available(self, message):
	self.published = self.return_address
	for q in self.queue:
		self.forward(q[0], self.published, q[1])
	self.queue = deque()
	return EXECUTING

def ProcessObject_EXECUTING_Subscribed(self, message):
	self.subscribed = message
	return EXECUTING

def ProcessObject_EXECUTING_Unknown(self, message):
	if self.published is None:
		message = cast_to(message, self.received_type)
		q = (message, self.return_address)
		self.queue.append(q)
		return EXECUTING
	self.forward(message, self.published, self.return_address)
	return EXECUTING

def ProcessObject_EXECUTING_Returned(self, message):
	if self.subscribed:
		clear_subscribed(self, self.subscribed)
	self.send(RemoveObject(self.object_address), PO.collector)

	# Wait thread has returned
	# Forward the result.
	code, page = message.message.code, message.message.page

	self.log(USER_TAG.ENDED, f'Process ({self.p.pid}) ended with {code}')

	if not page:
		output = None
	else:
		encoding = CodecJson()
		try:
			output = encoding.decode(page, Any())
		except CodecError as e:
			s = str(e)
			self.complete(Faulted(f'cannot decode output ({s}) not a standard executable?'))

	# If 0 then everything working according to plans. Also
	# catch the case where the framework sets a special non-zero
	# exit status for faults in CLI scenarios.
	if code == 0 or (code == FAULTY_EXIT and isinstance(output, Faulted)):
		self.complete(output)

	# This is outside normal framework operation.
	t = '<empty>' if output is None else output.__class__.__name__
	p = '<empty>' if len(page) < 1 else page[:32]
	self.complete(Faulted(f'non-standard process exit - code={code}, output={t}, page="{p}..."'))

def ProcessObject_EXECUTING_Stop(self, message):
	pid = self.p.pid
	try:
		os.kill(pid, signal.SIGINT)
	except OSError as e:
		self.complete(Faulted(f'cannot relay local Stop to "{pid}" as SIGINT'))
	return CLEARING

def ProcessObject_CLEARING_Stop(self, message):
	# Additional redundant hint. Ignored.
	return CLEARING

def ProcessObject_CLEARING_Returned(self, message):
	if self.subscribed:
		clear_subscribed(self, self.subscribed)
	self.send(RemoveObject(self.object_address), PO.collector)

	# Wait thread has returned
	# Forward the result.
	code, page = message.message.code, message.message.page

	self.log(USER_TAG.ENDED, f'Process aborted [{self.p.pid}] (code {code})')
	self.complete(Aborted())

PROCESS_DISPATCH = {
	INITIAL: (
		(Start,),
		()
	),
	PENDING: (
		(HostPort, Unknown),
		()
	),
	EXECUTING: (
		(Available, Subscribed, Unknown, Returned, Stop),
		()
	),
	CLEARING: (
		(Returned, Stop),
		()
	),
}

bind_statemachine(ProcessObject, PROCESS_DISPATCH, thread='process-object')

#
#
class Punctuation(object):
	"""A collection of strings for custom decoration of a command line.

	:param dash: string to place before a short-form flag
	:type dash: str
	:param double_dash: string to place before a long-form name
	:type double_dash: str
	:param list_ends: left-end and right-end characters bounding a list
	:type list_ends: str, len of 2
	:param list_separator: string to place between list elements
	:type list_separator: str
	:param dict_ends: left-end and right-end characters bounding a dict
	:type dict_ends: str, len of 2
	:param dict_separator: string to place between dict elements
	:type dict_separator: str
	:param dict_colon: str to place between name and value of dict pair
	:type dict_colon: str
	:param message_ends: left-end and right-end characters bounding a message
	:type message_ends: str, len of 2
	:param message_separator: string to place between message elements
	:type message_separator: str
	:param message_colon: str to place between name and value of dict pair
	:type message_colon: str
	:param true_false: strings to encode as representations for true and false
	:type true_false: list of 2 str
	:param no_value: string to encode as a None value
	:type no_value: str
	:param flag_value_separator: string to place between flag and value
	:type flag_value_separator: str
	:param any_separator: string to place between elements of an Any representation
	:type any_separator: str
	"""
	def __init__(self, dash=None, double_dash=None,
			list_ends=None, list_separator=None,
			dict_ends=None, dict_separator=None, dict_colon=None,
			message_ends=None, message_separator=None, message_colon=None,
			true_false=None, no_value=None,
			flag_value_separator=None, any_separator=None):
		self.dash = dash or '-'
		self.double_dash = double_dash or '--'
		self.list_ends = list_ends or [None, None]
		self.list_separator = list_separator or ','
		self.dict_ends = dict_ends or [None, None]
		self.dict_separator = dict_separator or ','
		self.dict_colon = dict_colon or ':'
		self.message_ends = message_ends or [None, None]
		self.message_separator = message_separator or ','
		self.message_colon = message_colon or ':'
		self.true_false = true_false or ['true', 'false']
		self.no_value = no_value or 'null'
		self.flag_value_separator = flag_value_separator or '='
		self.any_separator = any_separator or '/'

class Utility(Point, StateMachine):
	"""An async proxy object that starts and manages a non-standard sub-process.

	The named executable is started and the machine waits for termination. If stdin
	is a ``str`` the contents are written to an input pipe. If stdout is ``str`` (i.e. the class)
	the object will return the text received on the output pipe, in the :class:`~.lifecycle.Completed`
	message.

	Parameters are passed from the calling process to the child process by translating the
	positional parameters according to a few rules;

	* Each parameter (i.e. ``args[i]``) should be a tuple where the first element is
	  the name of the parameter and the second element is the runtime value of that name.
	* A 3-tuple can also be used where the middle element contains the separator to used
	  on the command line, between the name and the value.
	* Values are Python values and these are encoded in a best-guess fashion, e.g. a Python
	  int will be converted to the proper sequence of digits. A Python str will be
	  passed verbatim.
	* Explicit type information can be passed in ``args_schema``. This is a name-type dict
	  where the type value is used to control the encoding process, e.g. a Python float
	  can be described as a ``ar.ClockTime`` and the float will be converted to a full ISO
	  format string on the command line.
	* By default the command line is decorated with dashes and equals signs. Passing a
	  :class:`~.processing.Punctuation` parameter takes explicit control over those decorations.

	:param name: name of the executable file
	:type name: str
	:param args: positional args
	:type args: tuple
	:param args_schema: explicit type information about args
	:type args_schema: dict of kipjak type expressions
	:param punctuation: override standard decoration of command line
	:type punctuation: :class:`~.processing.Punctuation`
	:param stdin: text to pass to child
	:type stdin: str or None
	:param stdout: type of expected output, e.g. str
	:type stdout: type
	:param stderr: type of expected output, e.g. str
	:type stderr: type
	:param text: nature of pipe content - text or binary
	:type text: bool
	:param encoding: control encoding of text, passed to ``Popen()``
	:type encoding: str
	:param errors: control encoding errors, passed to ``Popen()``
	:type errors: str
	:param cwd: where to locate the sub-process
	:type cwd: str
	:param kw: additional parameters passed to ``Popen()``
	:type kw: named parameters dict
	"""
	def __init__(self, name, *args,
			args_schema=None, punctuation=None,
			stdin=None, stdout=None, stderr=None,
			text=False, encoding=None, errors=None,
			cwd=None,
			**kw):
		Point.__init__(self)
		StateMachine.__init__(self, INITIAL)
		self.name = name
		self.args = args
		self.args_schema = args_schema
		self.punctuation = punctuation or Punctuation()
		self.stdin = stdin
		self.stdout = stdout
		self.stderr = stderr
		self.text = text
		self.encoding = encoding
		self.errors = errors
		self.cwd = cwd
		self.input = None
		self.piping = False
		self.kw = kw
		self.p = None

def Utility_INITIAL_Start(self, message):
	# If no home has been loaded, path will resolve to
	# none, i.e. the default.
	executable = shutil.which(self.name)
	if executable is None:
		cwd = os.getcwd()
		self.complete(Faulted(f'cannot resolve executable "{self.name}" from "{cwd}"'))

	try:
		args = process_args(self.args, self.args_schema, self.punctuation)
	except ValueError as e:
		s = str(e)
		self.complete(Faulted(f'cannot process arguments for "{self.name}", {s}'))

	# Pipe work
	# 1. ACTION - nothing in, nothing out (default)
	# 2. SINK - something in, nothing out
	# 3. SOURCE - nothing in, something out
	# 4. FILTER - something in, something out

	stdin = self.stdin
	if isinstance(stdin, str):
		self.input = stdin
		self.stdin = PIPE
		self.text = True
	elif isinstance(stdin, (bytes, bytearray)):	# Block
		self.input = stdin
		self.stdin = PIPE
		self.text = False

	stdout = self.stdout
	if stdout == str:	 # Unicode
		if stdin and not self.text:
			raise ValueError('cannot support different input/output/error pipes')
		self.stdout = PIPE
		self.text = True
	elif stdout in (bytes, bytearray):	# Block
		if stdin and self.text:
			raise ValueError('cannot support different input/output/error pipes')
		self.stdout = PIPE
		self.text = False

	stderr = self.stderr
	if stderr == str:	 # Unicode
		if (stdin or stdout) and not self.text:
			raise ValueError('cannot support different input/output/error pipes')
		self.stderr = PIPE
		self.text = True
	elif stderr in (bytes, bytearray):	# Block
		if (stdin or stdout) and self.text:
			raise ValueError('cannot support different input/output pipes')
		self.stderr = PIPE
		self.text = False

	self.piping = stdin or stdout or stderr

	line = [executable]
	line.extend(args)

	c = ' '.join(line)
	self.console(c)
	try:
		start_new_session = CL.child_process
		self.p = Popen(line,
			start_new_session=True,
			stdin=self.stdin, stdout=self.stdout, stderr=sys.stderr,
			text=True, encoding='utf-8', errors='strict')
		#**self.kw)
	except OSError as e:
		s = f'cannot start process "{self.module_path}" ({e})'
		self.complete(Faulted(s))

	self.create(wait, self.p, True)

	return EXECUTING

def Utility_EXECUTING_Returned(self, message):
	code, page = message.message.code, message.message.page

	if code == 0:
		if not page:
			self.complete(None)
		output = cast_to(page, bytes_type)
		self.complete(output)
	self.complete(Faulted(f'child exit code {code}', 'expecting 0 (zero)'))

def Utility_EXECUTING_Stop(self, message):
	self.p.terminate()
	return CLEARING

def Utility_CLEARING_Returned(self, message):
	self.complete(Aborted())

UTILITY_DISPATCH = {
	INITIAL: (
		(Start,),
		()
	),
	EXECUTING: (
		(Returned, Stop),
		()
	),
	CLEARING: (
		(Returned,),
		()
	),
}

bind_statemachine(Utility, UTILITY_DISPATCH)


#
#
NoneType = type(None)

def write_if(r, s):
	if s:
		r.write(s)

def write_if_else(r, b, ie):
	if b:
		r.write(ie[0])
	else:
		r.write(ie[1])

def dash_style(name, punctuation):
	if len(name) == 1:
		return punctuation.dash
	return punctuation.double_dash

def resolve(name, value, schema, punctuation):
	if value is None:
		return None
	return str(value)

def process_args(args, schema, punctuation=None):
	punctuation = punctuation or Punctuation()
	line = []
	for i, a in enumerate(args):
		if isinstance(a, tuple):
			n = len(a)
			if not (n in [2, 3]):
				raise ValueError('tuple flag [%d] with unexpected length %d' % (i, n))

			name = a[0]
			if not isinstance(name, str):
				raise ValueError('tuple flag [%d] with strange name %r' % (i, name))

			separator = punctuation.flag_value_separator
			dash = dash_style(name, punctuation)
			if len(a) == 2:
				value = resolve(name, a[1], schema, punctuation)
			else:
				separator = a[1]
				value = resolve(name, a[2], schema, punctuation)

			if value is None:
				line.append('%s%s' % (dash, name))
			elif separator is None:
				line.append('%s%s' % (dash, name))
				line.append('%s' % (value,))
			else:
				line.append('%s%s%s%s' % (dash, name, separator, value))
		else:
			value = resolve(i, a, schema, punctuation)
			line.append('%s' % (value,))
	return line



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Derive the proper signal handling for the current host.

"""
__docformat__ = 'restructuredtext'

import os
import platform
import signal
from .virtual_runtime import *
from .general_purpose import *
from .command_line import *
from .object_runtime import *

__all__ = [
	'PS',
]

# Construct platform-independent materials here. This is
# intentionally low ambition. Test rigs for multi-os dev
# just not practical.
# Slightly confusing set of criteria available. Save them
# all for diagnostic but base operation on one.
def catch_interrupt(number, frame):
	PS.signal_received = number

def interrupt_alias(number, frame):
	root = start_up(None)
	root.log(USER_TAG.TRACE, f'Accepting signal {number} as SIGINT alias')
	PS.signal_received = signal.SIGINT

def ignore_signal(number, frame):
	pass

def log_signal(number, frame):
	root = start_up(None)
	# Skip the filtering.
	root.log(USER_TAG.WARNING, 'Unexpected signal {number}'.format(number=number))

system = platform.system()

#
PS = Gas(os_name=os.name,
	platform_release=platform.release(),
	platform_system=system,
	platform_signal=None,
	platform_kill=None,
	signal_received=None)

if system.startswith('Linux'):
	def platform_signal():
		signal.signal(signal.SIGINT, catch_interrupt)
		signal.signal(signal.SIGQUIT, interrupt_alias)
		signal.signal(signal.SIGHUP, interrupt_alias)
		signal.signal(signal.SIGTERM, interrupt_alias)

		signal.signal(signal.SIGCHLD, ignore_signal)
		signal.signal(signal.SIGTRAP, log_signal)
		signal.signal(signal.SIGABRT, log_signal)
		#signal.signal(signal.SIGKILL, log_signal)	... cant be caught.

		signal.signal(signal.SIGPIPE, log_signal)
		signal.signal(signal.SIGUSR1, catch_interrupt)
		signal.signal(signal.SIGUSR2, catch_interrupt)
		signal.signal(signal.SIGALRM, log_signal)
		signal.signal(signal.SIGTTIN, log_signal)
		#signal.signal(signal.SIGSTOP, log_signal)	... ditto.
		signal.signal(signal.SIGTSTP, log_signal)
		signal.signal(signal.SIGPWR, log_signal)
	PS.platform_kill = signal.SIGKILL
	PS.platform_signal = platform_signal
elif system == 'Darwin':
	def platform_signal():
		signal.signal(signal.SIGINT, catch_interrupt)
		signal.signal(signal.SIGQUIT, interrupt_alias)
		signal.signal(signal.SIGHUP, interrupt_alias)
		signal.signal(signal.SIGTERM, interrupt_alias)

		signal.signal(signal.SIGCHLD, ignore_signal)
		signal.signal(signal.SIGTRAP, log_signal)
		signal.signal(signal.SIGABRT, log_signal)
		#signal.signal(signal.SIGKILL, log_signal)	... cant be caught.

		signal.signal(signal.SIGPIPE, log_signal)
		signal.signal(signal.SIGUSR1, catch_interrupt)
		signal.signal(signal.SIGUSR2, catch_interrupt)
		signal.signal(signal.SIGALRM, log_signal)
		signal.signal(signal.SIGTTIN, log_signal)
		#signal.signal(signal.SIGSTOP, log_signal)	... ditto.
		signal.signal(signal.SIGTSTP, log_signal)
		#signal.signal(signal.SIGPWR, log_signal)
	PS.platform_kill = signal.SIGKILL
	PS.platform_signal = platform_signal
elif system == 'Windows':
	def platform_signal():
		signal.signal(signal.SIGINT, catch_interrupt)
		# Available but unusable. Result in termination of process.
		#signal.signal(signal.SIGTERM, interrupt_alias)
		#signal.signal(signal.SIGABRT, log_signal)
	PS.platform_kill = signal.SIGBREAK
	PS.platform_signal = platform_signal
else:
	pass



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Directory at the LAN scope.

Run the pub-sub name service at the LAN level.
"""
__docformat__ = 'restructuredtext'

import kipjak as kj

#
#
class INITIAL: pass
class RUNNING: pass
class CLEARING: pass

class Lan(kj.Threaded, kj.StateMachine):
	def __init__(self, directory_at_lan: kj.HostPort=None, encrypted_directory: bool=None):
		kj.Threaded.__init__(self)
		kj.StateMachine.__init__(self, INITIAL)
		self.directory_at_lan = directory_at_lan
		self.encrypted_directory = encrypted_directory

def Lan_INITIAL_Start(self, message):
	# Does not try to infer listening IP. Its either an argument or its all interfaces.
	accept_directories_at = self.directory_at_lan or kj.HostPort('0.0.0.0', kj.DIRECTORY_PORT)

	self.directory = self.create(kj.ObjectDirectory, directory_scope=kj.ScopeOfDirectory.LAN,
		accept_directories_at=accept_directories_at,
		encrypted=self.encrypted_directory)

	self.assign(self.directory, 1)
	return RUNNING

def Lan_RUNNING_Returned(self, message):
	f = kj.Faulted('directory terminated')
	self.complete(f)

def Lan_RUNNING_Faulted(self, message):
	self.abort(message)
	return CLEARING

def Lan_RUNNING_Stop(self, message):
	self.abort(kj.Aborted())
	return CLEARING

def Lan_CLEARING_Returned(self, message):
	self.complete(self.aborted_message)

LAN_DISPATCH = {
	INITIAL: (
		(kj.Start,),
		()
	),
	RUNNING: (
		(kj.Returned, kj.Faulted, kj.Stop),
		()
	),
	CLEARING: (
		(kj.Returned,),
		()
	),
}

kj.bind(Lan, LAN_DISPATCH)

# For package scripting.
def main():
	kj.create(Lan)

if __name__ == '__main__':
	main()



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Status and control settings bound to objects and messages.

The flags, enumerations and other values used to control handling
and behaviour and a class to hold this collection of administrative
details.

.. autoclass:: Runtime
"""

__docformat__ = 'restructuredtext'

from enum import Enum

__all__ = [
	'USER_LOG',
	'USER_TAG',
	'tag_to_log',
	'TAG_LOG',
	'Runtime',
	'type_schema',
	'PointConstructionError',
	'PlatformRequirementError',
]

# Logging levels. Moderate the quantity
# of logging by its designated significance.
class USER_LOG(Enum):
	"""
	Enumeration of the levels within the **kipjak** logging.

	* NONE - no logs at all
	* FAULT - a definite problem that will compromise the service.
	* WARNING - something unexpected that may compromise the service.
	* CONSOLE - an operational milestone worthy of note.
	* OBJECT - asynchronous operation
	* TRACE - progress of the service, suitable for public viewing.
	* DEBUG - not suitable for customer or support.
	"""
	NONE = 100
	FAULT = 6
	WARNING = 5
	CONSOLE = 4
	OBJECT = 3
	TRACE = 2
	DEBUG = 1

# Async operation.
class USER_TAG(Enum):
	"""
	Enumeration of the tags used to stamp each log entry.

	* CREATED ``+`` new object
	* DESTROYED ``X`` object gone
	* SENT ``>`` object sent a message
	* RECEIVED ``<`` object received a message
	* STARTED ``(`` object started a subprocess
	* ENDED ``)`` subprocess ended
	* FAULT ``!`` compromised operation
	* WARNING ``?`` may be compromised
	* CONSOLE ``^`` application milestone
	* TRACE ``~`` technical notes
	* DEBUG ``_`` developer notes
	* SAMPLE ``&`` formal sample of local data
	* CHECK ``=`` assert a condition
	"""
	CREATED = '+'
	DESTROYED = 'X'
	SENT = '>'
	RECEIVED = '<'
	STARTED = '('
	ENDED = ')'

	# Operational significance.
	FAULT = '!'
	WARNING = '?'
	CONSOLE = '^'

	TRACE = '~'
	DEBUG = '_'

	SAMPLE = '&'
	CHECK = '='

TAG_LOG = {
	USER_TAG.FAULT.value: USER_LOG.FAULT,
	USER_TAG.WARNING.value: USER_LOG.WARNING, USER_TAG.CHECK.value: USER_LOG.WARNING,
	USER_TAG.CONSOLE.value: USER_LOG.CONSOLE,
	USER_TAG.SAMPLE.value: USER_LOG.TRACE,
	USER_TAG.CREATED.value: USER_LOG.OBJECT, USER_TAG.DESTROYED.value: USER_LOG.OBJECT,
	USER_TAG.SENT.value: USER_LOG.OBJECT, USER_TAG.RECEIVED.value: USER_LOG.OBJECT,
	USER_TAG.STARTED.value: USER_LOG.OBJECT, USER_TAG.ENDED.value: USER_LOG.OBJECT,
	USER_TAG.TRACE.value: USER_LOG.TRACE,
	USER_TAG.DEBUG.value: USER_LOG.DEBUG
}

def tag_to_log(tag):
	"""Convert tag to level. Return int."""
	number = TAG_LOG[tag.value]
	return number

class Runtime(object):
	"""Settings to control logging and other behaviour, for objects and messages."""

	def __init__(self,
			name, module,
			lifecycle=True, message_trail=True,
			execution_trace=True,
			copy_before_sending=True,
			not_portable=False,
			user_logs=USER_LOG.DEBUG):
		"""Construct the settings.

		:param name: the name of the class being registered
		:type name: str
		:param module: the name of the module the class is located in
		:type module: str
		:param lifecycle: enable logging of created, destroyed
		:type lifecycle: bool
		:param message_trail: enable logging of sent
		:type message_trail: bool
		:param execution_trace: enable logging of received
		:type execution_trace: bool
		:param copy_before_sending: enable auto-copy before send
		:type copy_before_sending: bool
		:param not_portable: prevent inappropriate send
		:type not_portable: bool
		:param user_logs: log level
		:type user_logs: int
		"""
		self.name = name		# Last component of dotted name.
		self.module = module	# Full path up to the name.
		self.schema = None

		self.lifecycle = lifecycle			  # Create, destroy objects
		self.message_trail = message_trail	  # Sending
		self.execution_trace = execution_trace  # Receiving
		self.copy_before_sending = copy_before_sending
		self.not_portable = not_portable
		self.user_logs = user_logs			  # Object trace, warning...

		self.path = f'{module}.{name}'

def type_schema(p):
	try:
		return p.__art__.schema
	except (TypeError, AttributeError) as e:
		return None

#
class PointConstructionError(Exception):
	"""Exception indicating poor construction of an async entity."""

	def __init__(self, identify_and_help):
		"""Construct the exception.

		:param identity_and_help: description of the problem and a suggestion
		:type name: str
		"""
		Exception.__init__(self, identify_and_help)

class PlatformRequirementError(Exception):
	"""Exception indicating that the underlying platform is not meeting its end of the deal."""

	def __init__(self, identify_and_help):
		"""Construct the exception.

		:param identity_and_help: the problem and a suggestion
		:type name: str
		"""
		Exception.__init__(self, identify_and_help)



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Management of the async runtime.

Ensure that the support for async operation is in place when the process
needs it. Ensure that support is cleared out during process termination.
"""
__docformat__ = 'restructuredtext'

import threading
import tempfile
from .general_purpose import *
from .virtual_runtime import *
from .command_line import *
from .folder_object import *

__all__ = [
	'HR',
	'resource_folder',
	'model_folder',
	'tmp_folder',
	'resource_path',
	'model_path',
	'tmp_path',
]

HR = Gas(home_path=None, role_name=None,
	home_role=None,
	edit_role=None,
	temp_dir=None,
	model=None, tmp=None, resource=None)

# File storage areas for an instance of a process object.
# Specifically deals with both group and solo (CLI) contexts.
# Resource - read-only, persistent, shared by all instances of same executable.
# Model - read-write, persistent, private to each instance.
# Tmp - read-write, empty on start, private.

# Set in create_role(), open_role() and create_memory_role().
def resource_folder():
	"""Application access to the shared, per-executable file space. Return a folder.

	Static, read-only file data associated with the running executable. This is
	intended to be the folder managed by the :ref:`resource<command-reference-resource>`
	command but may be explicitly set on the command-line using ``--resource-path``, or it
	is ``None``.

	:rtype: Folder
	"""
	return HR.resource

def model_folder():
	"""Application access to the operational, per-role file space. Return a folder.

	Private, persistent, read-write file data associated with the operational role. This is
	intended to be the folder managed by the :ref:`model<command-reference-model>`
	command but may be explicitly set on the command-line using ``--model-path``, or it
	is the current working folder.

	:rtype: Folder
	"""
	return HR.model

tmp_lock = threading.RLock()

def tmp_folder():
	"""Application access to the transient, per-role file space. Return a folder.

	Private, transient, read-write file data associated with the operational *role*. This is
	a space arranged by the framework, where temporary files can be freely added, modified
	and deleted. It is part of the *home* folder hierarchy, or - when the process is running
	from the command-line - it is a folder created within the folders managed by the host
	platform. The folder is guaranteed to be empty at the startup of every application.

	:rtype: Folder
	"""
	with tmp_lock:
		if not HR.tmp:
			# Cleanup in create()
			HR.temp_dir = tempfile.TemporaryDirectory()
			HR.tmp = Folder(HR.temp_dir.name)
	return HR.tmp

def resource_path():
	"""Application access to the shared, per-executable file space. Return the path or None.

	:rtype: str
	"""
	f = resource_folder()
	if isinstance(f, Folder):
		return f.path
	return None

def model_path():
	"""Application access to the operational, per-role file space. Return the path.

	:rtype: str
	"""
	f = model_folder()
	if isinstance(f, Folder):
		return f.path
	return None

def tmp_path():
	"""Application access to the transient, per-role file space. Return the path.

	:rtype: str
	"""
	f = tmp_folder()
	if isinstance(f, Folder):
		return f.path
	return None



# test_main_args.py
import uuid
import kipjak as kj

from test_person import *

__all__ = [
	'Main',
	'Main_Start',
	'Main_int',
	'Main_list_int',
	'Main_Person',
	'Main_dict_UUID_Person',
	'Main_Stop',
]

class Main(kj.Point, kj.Stateless):
	def __init__(self, height: int=8, width: int=8, value: float=0.125):
		super().__init__()
		self.height = height
		self.width = width
		self.value = value

def Main_Start(self, message):
	self.console(height=self.height, width=self.width, value=self.value)

def Main_int(self, message):
	self.console(message=message)

def Main_list_int(self, message):
	self.console(message=message)

def Main_Person(self, message):
	self.console(message=message)

def Main_dict_UUID_Person(self, message):
	self.console(message=message)

def Main_Stop(self, message):
	table = [[self.value] * self.height] * self.width
	self.complete(table)

kj.bind(Main,
	(kj.Start, int, list[int], Person, dict[uuid.UUID, Person], kj.Stop),
	return_type=kj.VectorOf(kj.VectorOf(float)))

if __name__ == '__main__':
	kj.create(Main)



#
#
import kipjak as kj

# MESSAGES ------------------------------------------------------------------------------
class Cook(object):
	def __init__(self, a: int=None, b: dict[int, str]=None, c=None):	# Cant hint here (-> int). Would annoy checkers.
		self.a = a
		self.b = b
		self.c = c

class Radiating(object):
	pass

kj.bind_message(Cook, c=kj.PointerTo(kj.Boolean))
kj.bind_message(Radiating)

# 1. List of args and their types and a return type, from function hints.
# 2. Still must be default callable (policy around persisted settings).
# 3. Types provided on bind when hints fall short.
'''
# POINT AS STATELESS MACHINE ------------------------------------------------------------
class DbQuery(kj.Point, kj.Stateless):
	def __init__(self, a: int=None, b: dict[int, str]=None, c=None):
		super().__init__()
		self.a = a
		self.b = b
		self.c = c

def DbQuery_Start(self, message):
	pass

def DbQuery_T1(self, message):
	pass

def DbQuery_Stop(self, message):
	self.complete()

kj.bind_stateless(DbQuery, (kj.Start, kj.T1, kj.Stop),
	return_type=dict[int, str],
	c=kj.TimeSpan())

# 1. List of args and their types, from __init__ function hints.
# 2. Dispatch table (list of received types) on bind.
# 3. Dispatch types should be hints where possible, Portable when not.
# 4. A return_type is passed to bind.
# 2. Still must be default callable (policy around persisted settings).
# 3. Types provided on bind when hints fall short.

# POINT AS FINITE-STATE MACHINE ---------------------------------------------------------
class INITIAL: pass
class IDLE: pass
class COOKING: pass

class Microwave(kj.Threaded, kj.StateMachine):
	def __init__(self, a: int=None, b: dict[int, str]=None, c=None):
		super().__init__(INITIAL)
		self.a = a
		self.b = b
		self.c = c
		self.temperature = 0
		self.timeout = 0

def Microwave_INITIAL_Start(self, message):
	return IDLE

def Microwave_IDLE_Cook(self, message):
	self.reply(Radiating())
	return COOKING

def Microwave_COOKING_Stop(self, message):
	self.complete(44)

kj.bind_stateful(Microwave,
	{
		INITIAL: (
			(kj.Start,),
			()
		),
		IDLE: (
			(Cook,),
			()
		),
		COOKING: (
			(kj.Stop,),
			()
		)
	},
	return_type=dict[int, str],
	c=kj.TimeSpan())


class Cook(object):
	def __init__(self, a: int=None, b: dict[int, str]=None, c=None):	# Cant hint here (-> int). Would annoy checkers.
kj.bind_message(Cook, c=kj.PointerTo(kj.Boolean))
BR_1 = kj.branch(int, dict[int, str], kj.Start)
def main(self, a: int=None, b: dict[int, str]=None, c=None) -> int:
	i, m = self.select(BR_1)
RT_2 = kj.register(dict[int, str])

def poly(self, a: int=None, b: dict[int, str]=None, c=None) -> kj.Any:	# Or typing.Any? NO - DIFFERENT
	b = b or {}

	isinstance(a, int)
	isinstance(b, dict)

	m, i = self.select(BR_1)
	if i == 0:
		return (0, RT_1)

	return ({'a':1}, RT_2)

class DbQuery(kj.Point, kj.Stateless):
	def __init__(self, a: int=None, b: dict[int, str]=None, c=None):

kj.bind_stateless(DbQuery, (kj.Start, kj.T1, kj.Stop),

class Microwave(kj.Threaded, kj.StateMachine):
	def __init__(self, a: int=None, b: dict[int, str]=None, c=None):
		super().__init__(INITIAL)

kj.bind_stateful(Microwave,
	{
		INITIAL: (
			(kj.Start,),
			()
		),

1. Message.__init__(self, a: int=None, b: dict[int, str]=None, c=None):
2. bind_message(Cook, c=kj.PointerTo(kj.Boolean)))
3. BR_1 = kj.branch(int, dict[int, str], kj.Start)
4. def main(self, a: int=None, b: dict[int, str]=None, c=None) -> int:
5. i, m = self.select(BR_1)
6. RT_2 = kj.register(dict[int, str])
7. def poly(self, a: int=None, b: dict[int, str]=None, c=None) -> kj.Any:
8. m, i = self.select(BR_1)
9. return ({'a':1}, RT_2)
10: class DbQuery(kj.Point, kj.Stateless): def __init__(self, a: int=None, b: dict[int, str]=None, c=None):
11: kj.bind_stateless(DbQuery, (kj.Start, kj.T1, kj.Stop), return_type=dict[str,int])
12: class Microwave(kj.Threaded, kj.StateMachine): def __init__(self, a: int=None, b: dict[int, str]=None, c=None):
13. kj.bind_stateful(Microwave,	{ INITIAL: ((kj.Start,), () ),


1. get_type_hints(Message.__init__) --------------------------- bind_message, hints to Portable, lookup/install
2. compile_schema(Message, c=kj.PointerTo(kj.Boolean))--------- compile_schema, if k in kw, override
3. def branch(*a)
4. get_type_hints(main) --------------------------------------- bind_function, hints to Portable, return type, lookup/install
5. BR_2 = kj.branch_table(dict[int, str]) --------------------- branch_table, hints to Portable, lookup/install, return tuple
6. i, m = self.branch(table-of-types) ------------------------- i, m = self.select(table), new input strategy for new type system
7. RT_2 = kj.return_type(dict[int, str]) ---------------------- RT = return_type(), hint to Portable, lookup/install
8. get_type_hints(Stateless.__init__) ------------------------- bind_stateless, hints to Portable, lookup/install
9. compile_schema(Stateless, c=kj.PointerTo(kj.Boolean)) ------ bind_stateless, return_type
10. get_type_hints(StateMachine.__init__) --------------------- bind_statemachine, hints to Portable, lookup/install
11. compile_schema(StateMachine, c=kj.PointerTo(kj.Boolean)) -- bind_statemachine, return_type
12. def main(self, b: dict[int, str]=None) -> int: ------------ argv -> codec -> settings, return -> codec -> stdout

13. self.select(int, Start, ...) ------------------------------ keep as easy-but-slow, for learning
'''



# Author: Scott Woods <kipjak.baker@gmail.com>
# MIT License
#
# Copyright (c) 2025 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Support for the concept of attempts and retries.

There is an object that tries to complete a task. If it fails, subsequent
attempts may be appropriate. A clear and robust implementation of this concept
is hugely useful in the context of networking. There is also a need for
refinement - "blunt force" retries can quickly create its own problems.
"""
__docformat__ = 'restructuredtext'

import random

from .bind_type import *

__all__ = [
	'RetryIntervals',
	'intervals_only',
	'smart_intervals',
]

#
#
random.seed()

# All times are subject to the accuracy of the internal timer service, e.g. 0.25s
# The randomized value should be a multiple of the timer accuracy.
# Step 1 should be at least twice of randomized
# The value of first steps should increase.
# Regular steps should be bigger than the last first step.
# All attributes except first_steps, can be set to None.
# There must be at least first steps or regular steps or both.
# Regular steps without a truncated value is an expression of forever.
# The defaults are based on the retry strategy from SIP.
# Randomization is loosely based on exponential backoff.

class RetryIntervals(object):
	"""
	Capture the values needed to manage exponential backoff.

	:param first_steps: initial series of explicit float values
	:param regular_steps: float value repeated after ``first_steps``
	:param step_limit: maximum number of steps
	:param randomized: micro-steps used when randomizing
	:param truncated: add step * truncated when randomizing
	"""
	def __init__(self, first_steps: list[float]=None, regular_steps: float=None, step_limit: int=None, randomized: float=None, truncated: float=None):
		self.first_steps = first_steps or []
		self.regular_steps = regular_steps
		self.step_limit = step_limit
		self.randomized = randomized
		self.truncated = truncated

	def empty(self):
		if len(self.first_steps) > 0:
			return False
		if self.regular_steps is not None:
			return False
		return True

bind(RetryIntervals)

# Coroutine to generate the sequence
# implied by first_steps and regular_steps.
def no_limit(retry):
	for i in retry.first_steps:
		yield i

	if retry.regular_steps is None:
		return

	while True:
		yield retry.regular_steps

def intervals_only(retry):
	limit = retry.step_limit
	for i in retry.first_steps:
		limit -= 1
		if limit < 1:
			return
		yield i
	if retry.regular_steps is None:
		return
	# For ever perhaps capped by
	# limit.
	while True:
		limit -= 1
		if limit < 1:
			return
		yield retry.regular_steps

# Wrap it in another coroutine that implements
# truncations and randomization of the underlying
# sequence.
def smart_intervals(retry):
	limit = retry.step_limit
	randomized = retry.randomized
	truncated = retry.truncated

	def cooked(r):
		if randomized is None:
			return r
		if truncated is None:
			t = r
		else:
			t = r * truncated
		m = int(t / retry.randomized) + 1
		s = random.randrange(0, m)
		c = s * retry.randomized
		return r + c

	if limit is None:
		for r in no_limit(retry):
			yield cooked(r)
		return

	for _, r in zip(range(limit), intervals_only(retry)):
		yield cooked(r)



# test_subscriber.py
import kipjak as kj

from test_api import *

# A bare-bones implementation of a client subscribing to a network service.
DEFAULT_SEARCH = 'acme'

def subscriber(self, search: str=None, enduring: bool=False) -> list[list[float]]:
	'''Open communications with a named service and exchange messages. Return table.'''
	search = search or DEFAULT_SEARCH

	kj.subscribe(self, search, scope=kj.ScopeOfDirectory.LAN)		# Connect this object with the named object.

	m = self.input()
	if isinstance(m, kj.Subscribed):		# Search registered with directory.
		pass

	elif isinstance(m, kj.Faulted):			# Any fault, e.g. NotSubscribed
		self.complete(m)

	# Run a live directory client. Framework notifications followed by
	# requests and responses.
	while True:
		m = self.input()

		if isinstance(m, kj.Available):		# Connected to the service.
			self.send(Xy(x=2, y=2), self.return_address)

		elif isinstance(m, list):			# Table.
			if enduring:
				continue
			self.complete(m)

		elif isinstance(m, kj.Dropped):		# Lost the server.
			continue

		elif isinstance(m, kj.Stop):
			self.complete(kj.Aborted())

kj.bind(subscriber)				# Register with framework.

if __name__ == '__main__':		# Process entry-point.
	kj.create(subscriber)



# test_echo.py
import kipjak as kj
from kipjak.listen_connect import *

import test_echo

def main(self, requested_ipp: kj.HostPort=None):
	requested_ipp = requested_ipp or kj.HostPort('127.0.0.1', 5010)

	self.console(requested_ipp=requested_ipp)

	echo = self.create(kj.ProcessObject, test_echo.main)

	self.start(kj.T1, 2.0)
	m, i = self.select(kj.T1, kj.Faulted, kj.Stop)

	connect(self, requested_ipp=kj.HostPort('127.0.0.1', 5010))
	m, i = self.select(Connected, kj.Faulted, kj.Stop)
	assert isinstance(m, Connected)
	server = self.return_address

	self.send(kj.Ack(), server)
	m, i = self.select(kj.Ack, kj.Faulted, kj.Stop)
	assert isinstance(m, kj.Ack)

	self.start(kj.T1, 5.0)
	self.select()

	self.send(Close(), server)
	m, i = self.select(Closed, kj.Faulted, kj.Stop)
	assert isinstance(m, Closed)

	self.send(kj.Stop(), echo)
	m, i = self.select(kj.Returned, kj.Faulted, kj.Stop)
	assert isinstance(m, kj.Returned)

kj.bind(main)

if __name__ == '__main__':
	kj.create(main)
